from __future__ import annotations

import json
from datetime import datetime
from pathlib import Path

import numpy as np

from openlifu.db.subject import Subject
from openlifu.geo import Point
from openlifu.plan.solution_analysis import SolutionAnalysisOptions
from openlifu.seg.material import Material
from openlifu.xdc.element import Element
from openlifu.xdc.transducer import Transducer


class PYFUSEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, Point):
            return obj.to_dict()
        if isinstance(obj, Transducer):
            return obj.to_dict()
        if isinstance(obj, Element):
            return obj.to_dict()
        if isinstance(obj, Material):
            return obj.to_dict()
        if isinstance(obj, Subject):
            return obj.to_dict()
        if isinstance(obj, SolutionAnalysisOptions):
            return obj.to_dict()
        return super().default(obj)

def to_json(obj, filename):
    dirname = Path(filename).parent
    if dirname and not dirname.exists():
        dirname.mkdir(parents=True)
    with open(filename, 'w') as file:
        json.dump(obj, file, cls=PYFUSEncoder, indent=4)



from __future__ import annotations

import logging
import os
import sys
import threading
import time
from pathlib import Path

if os.name == 'nt':
    import msvcrt
else:
    import select

import numpy as np

from openlifu.bf.pulse import Pulse
from openlifu.bf.sequence import Sequence
from openlifu.db import Database
from openlifu.geo import Point
from openlifu.io.LIFUInterface import LIFUInterface
from openlifu.plan.solution import Solution

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_watertank.py

"""
Test script to automate:
1. Connect to the device.
2. Test HVController: Turn HV on/off and check voltage.
3. Test Device functionality.
"""

# Configure logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Prevent duplicate handlers and cluttered terminal output
if not logger.hasHandlers():
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))
    logger.addHandler(handler)
    logger.propagate = False

log_interval = 1  # seconds; you can adjust this variable as needed

# set focus
xInput = 0
yInput = 0
zInput = 50

frequency_kHz = 400 # Frequency in kHz
voltage = 10.0 # Voltage in Volts
duration_msec = 5 # Pulse Duration in milliseconds
interval_msec = 100 # Pulse Repetition Interval in milliseconds
num_modules = 2 # Number of modules in the system

use_external_power_supply = True # Select whether to use console or power supply

console_shutoff_temp_C = 70.0 # Console shutoff temperature in Celsius
tx_shutoff_temp_C = 70.0 # TX device shutoff temperature in Celsius
ambient_shutoff_temp_C = 70.0 # Ambient shutoff temperature in Celsius

#TODO: script_timeout_minutes = 30 # Prevent unintentionally leaving unit on for too long
#TODO: log_temp_to_csv_file = True # Log readings to only terminal or both terminal and CSV file

# Fail-safe parameters if the temperature jumps too fast
rapid_temp_shutoff_C = 40 # Cutoff temperature in Celsius if it jumps too fast
rapid_temp_shutoff_seconds = 5 # Time in seconds to reach rapid temperature shutoff
rapid_temp_increase_per_second_shutoff_C = 3 # Rapid temperature climbing shutoff in Celsius

peak_to_peak_voltage = voltage * 2 # Peak to peak voltage for the pulse

here = Path(__file__).parent.resolve()
db_path = here / ".." / "db_dvc"
db = Database(db_path)
arr = db.load_transducer(f"openlifu_{num_modules}x400_evt1")
arr.sort_by_pin()


target = Point(position=(xInput,yInput,zInput), units="mm")
focus = target.get_position(units="mm")
distances = np.sqrt(np.sum((focus - arr.get_positions(units="mm"))**2, 1)).reshape(1,-1)
tof = distances*1e-3 / 1500
delays = tof.max() - tof
#delays = delays*0.0

apodizations = np.ones((1, arr.numelements()))
#active_element = 25
#active_element = np.arange(1,65)
#apodizations = np.zeros((1, arr.numelements()))
#apodizations[:, active_element-1] = 1


logger.info("Starting LIFU Test Script...")
interface = LIFUInterface(ext_power_supply=use_external_power_supply)
tx_connected, hv_connected = interface.is_device_connected()

if not use_external_power_supply and not tx_connected:
    logger.warning("TX device not connected. Attempting to turn on 12V...")
    interface.hvcontroller.turn_12v_on()

    # Give time for the TX device to power up and enumerate over USB
    time.sleep(2)

    # Cleanup and recreate interface to reinitialize USB devices
    interface.stop_monitoring()
    del interface
    time.sleep(1)  # Short delay before recreating

    logger.info("Reinitializing LIFU interface after powering 12V...")
    interface =  LIFUInterface(ext_power_supply=use_external_power_supply)

    # Re-check connection
    tx_connected, hv_connected = interface.is_device_connected()

if not use_external_power_supply:
    if hv_connected:
        logger.info(f"  HV Connected: {hv_connected}")
    else:
        logger.error("❌ HV NOT fully connected.")
        sys.exit(1)
else:
    logger.info("  Using external power supply")

if tx_connected:
    logger.info(f"  TX Connected: {tx_connected}")
    logger.info("✅ LIFU Device fully connected.")
else:
    logger.error("❌ TX NOT fully connected.")
    sys.exit(1)

stop_logging = False  # flag to signal the logging thread to stop

def log_temperature():
    # Create a file with the current timestamp in the name
    start = time.time()
    timestamp = time.strftime("%Y%m%d_%H%M%S")
    logpath = here / ".." / "logs"
    if not logpath.exists():
        logpath.mkdir(parents=True)
    logfile = logpath / f"{timestamp}_{frequency_kHz}kHz_{voltage}V_{duration_msec}ms_Duration_{interval_msec}ms_Interval_Temperature_Readings.csv"
    shutdown = False

    prev_tx_temp = None
    prev_amb_temp = None
    prev_con_temp = None

    if use_external_power_supply:
        con_temp = "N/A"

    with open(logfile, "w") as logfile:
        # Create header for CSV file
        log_line = "Current Time,Frequency (kHz),Duration (ms),Interval (ms),Voltage (Per Rail),Voltage (Peak to Peak),Console Temperature (°C),Transmitter Temperature (°C),Ambient Temperature (°C)\n"
        logfile.write(log_line)
        logfile.flush()  # Ensure the data is written immediately
        while not (stop_logging or shutdown):
            current_time = time.strftime("%Y-%m-%d %H:%M:%S")

            # Timer for initial rapid temperature increase
            within_initial_time_threshold = (time.time() - start) < rapid_temp_shutoff_seconds

            ## Check for too high of a temperature increase between readings
            # Console general temperature increase (bypass if using external power supply)
            if not use_external_power_supply:
                if prev_con_temp is None:
                    prev_con_temp = interface.hvcontroller.get_temperature1()
                con_temp = interface.hvcontroller.get_temperature1()
                if (con_temp - prev_con_temp) > rapid_temp_increase_per_second_shutoff_C:
                    logger.warning(f"Console temperature rose from {prev_con_temp}°C to {con_temp}°C (above {rapid_temp_increase_per_second_shutoff_C}°C threshold) within {log_interval}s.")
                    log_line = f"{current_time},SHUTDOWN,Console temperature exceeded rapid temp increase shutoff threshold\n"
                    shutdown=True
                else:
                    prev_con_temp = con_temp

            # TX device general temperature increase
            if prev_tx_temp is None:
                prev_tx_temp = interface.txdevice.get_temperature()
            tx_temp = interface.txdevice.get_temperature()
            if (tx_temp - prev_tx_temp) > rapid_temp_increase_per_second_shutoff_C:
                logger.warning(f"TX device temperature rose from {prev_tx_temp}°C to {tx_temp}°C (above {rapid_temp_increase_per_second_shutoff_C}°C threshold) within {log_interval}s.")
                log_line = f"{current_time},SHUTDOWN,TX device temperature exceeded rapid temp increase shutoff threshold\n"
                shutdown=True
            else:
                prev_tx_temp = tx_temp

            # Ambient temperature general increase
            if prev_amb_temp is None:
                prev_amb_temp = interface.txdevice.get_ambient_temperature()
            amb_temp = interface.txdevice.get_ambient_temperature()
            if (amb_temp - prev_amb_temp) > rapid_temp_increase_per_second_shutoff_C:
                logger.warning(f"Ambient temperature rose from {prev_amb_temp}°C to {amb_temp}°C (above {rapid_temp_increase_per_second_shutoff_C}°C threshold) within {log_interval}s.")
                log_line = f"{current_time},SHUTDOWN,Ambient temperature exceeded rapid temp increase shutoff threshold\n"
                shutdown=True
            else:
                prev_amb_temp = amb_temp

            # Check for initial rapid temperature increase
            if (within_initial_time_threshold):
                if not use_external_power_supply and (con_temp > rapid_temp_shutoff_C):
                    logger.warning(f"Console temperature {con_temp}°C exceeds rapid shutoff threshold of {rapid_temp_shutoff_C}°C within {rapid_temp_shutoff_seconds}s.")
                    log_line = f"{current_time},SHUTDOWN,Console temperature exceeded rapid shutoff threshold\n"
                    shutdown=True
                elif (tx_temp > rapid_temp_shutoff_C):
                    logger.warning(f"TX device temperature {tx_temp}°C exceeds rapid shutoff threshold of {rapid_temp_shutoff_C}°C within {rapid_temp_shutoff_seconds}s.")
                    log_line = f"{current_time},SHUTDOWN,TX device temperature exceeded shutoff threshold\n"
                    shutdown=True
                elif (amb_temp > rapid_temp_shutoff_C):
                    logger.warning(f"Ambient temperature {amb_temp}°C exceeds rapid shutoff threshold of {rapid_temp_shutoff_C}°C within {rapid_temp_shutoff_seconds}s.")
                    log_line = f"{current_time},SHUTDOWN,Ambient temperature exceeded shutoff threshold\n"
                    shutdown=True

            log_line = f"{current_time},{frequency_kHz},{duration_msec},{interval_msec},{voltage},{peak_to_peak_voltage},{con_temp},{tx_temp},{amb_temp}\n"
            if not use_external_power_supply:
                logger.info(f"Console Temp: {con_temp}°C TX Temp: {tx_temp}°C Ambient Temp: {amb_temp}°C")
            else:
                logger.info(f"Console Temp: {con_temp} TX Temp: {tx_temp}°C Ambient Temp: {amb_temp}°C")
            logfile.write(log_line)
            logfile.flush()  # Ensure the data is written immediately

            # Check if any temperature exceeds the shutoff threshold
            if not use_external_power_supply and (con_temp > console_shutoff_temp_C):
                logger.warning(f"Console temperature {con_temp}°C exceeds shutoff threshold {console_shutoff_temp_C}°C.")
                log_line = f"{current_time},SHUTDOWN,Console temperature exceeded shutoff threshold\n"
                shutdown=True
            if tx_temp > tx_shutoff_temp_C:
                logger.warning(f"TX device temperature {tx_temp}°C exceeds shutoff threshold {tx_shutoff_temp_C}°C.")
                log_line = f"{current_time},SHUTDOWN,TX device temperature exceeded shutoff threshold\n"
                shutdown=True
            if amb_temp > ambient_shutoff_temp_C:
                logger.warning(f"Ambient temperature {amb_temp}°C exceeds shutoff threshold {ambient_shutoff_temp_C}°C.")
                log_line = f"{current_time},SHUTDOWN,Ambient temperature exceeded shutoff threshold\n"
                shutdown=True

            if shutdown:
                interface.txdevice.stop_trigger()
                logfile.write(log_line)
                logfile.flush()  # Ensure the data is written immediately
                break
            time.sleep(log_interval)
    minutes, seconds = divmod(int(time.time() - start), 60)
    logger.info(f"Temperature logging stopped after {minutes}:{seconds:02d}.")
    logger.info(f"Data saved to \"{os.path.relpath(logfile.name)}\".")
    sys.exit(0)

# Verify communication with the devices
if not interface.txdevice.ping():
    logger.error("Failed to ping the transmitter device.")
    sys.exit(1)

if not use_external_power_supply and not interface.hvcontroller.ping():
    logger.error("Failed to ping the console device.")
    sys.exit(1)

if not use_external_power_supply:
    try:
        console_firmware_version = interface.hvcontroller.get_version()
        logger.info(f"Console Firmware Version: {console_firmware_version}")
    except Exception as e:
        logger.error(f"Error querying console firmware version: {e}")
try:
    tx_firmware_version = interface.txdevice.get_version()
    logger.info(f"TX Firmware Version: {tx_firmware_version}")
except Exception as e:
    logger.error(f"Error querying TX firmware version: {e}")

logger.info("Enumerate TX7332 chips")
num_tx_devices = interface.txdevice.enum_tx7332_devices()
if num_tx_devices == 0:
    raise ValueError("No TX7332 devices found.")
elif num_tx_devices == num_modules*2:
    logger.info(f"Number of TX7332 devices found: {num_tx_devices}")
    numelements = 32*num_tx_devices
else:
    raise Exception(f"Number of TX7332 devices found: {num_tx_devices} != 2x{num_modules}")

logger.info(f'Apodizations: {apodizations}')
logger.info(f'Delays: {delays}')

pulse = Pulse(frequency=frequency_kHz*1e3, duration=duration_msec*1e-3)

sequence = Sequence(
    pulse_interval=interval_msec*1e-3,
    pulse_count=int(60/(interval_msec*1e-3)),
    pulse_train_interval=0,
    pulse_train_count=1
)

pin_order = np.argsort([el.pin for el in arr.elements])
solution = Solution(
    delays = delays[:, pin_order],
    apodizations = apodizations[:, pin_order],
    transducer=arr,
    pulse = pulse,
    voltage=voltage,
    sequence = sequence
)
profile_index = 1
profile_increment = True
trigger_mode = "continuous"

if use_external_power_supply:
    logger.info(f"Using external power supply. Ensure HV is turned on and set to {voltage}V before starting the trigger.")
interface.set_solution(
    solution=solution,
    profile_index=profile_index,
    profile_increment=profile_increment,
    trigger_mode=trigger_mode)

logger.info("Get Trigger")
trigger_setting = interface.txdevice.get_trigger_json()
if trigger_setting:
    logger.info(f"Trigger Setting: {trigger_setting}")
else:
    logger.error("Failed to get trigger setting.")
    sys.exit(1)


duty_cycle = int((duration_msec/interval_msec) * 100)
if duty_cycle > 50:
    logger.warning("❗❗ Duty cycle is above 50% ❗❗")

logger.info(f"User parameters set: \n\
    Module Invert: {arr.module_invert}\n\
    Frequency: {frequency_kHz}kHz\n\
    Voltage Per Rail: {voltage}V\n\
    Voltage Peak to Peak: {peak_to_peak_voltage}V\n\
    Duration: {duration_msec}ms\n\
    Interval: {interval_msec}ms\n\
    Duty Cycle: {duty_cycle}%\n\
    Use External Power Supply: {use_external_power_supply}\n\
    Initial Temp Safety Shutoff: Increase to {rapid_temp_shutoff_C}°C within {rapid_temp_shutoff_seconds}s of starting.\n\
    General Temp Safety Shutoff: Increase of {rapid_temp_increase_per_second_shutoff_C}°C within {log_interval}s at any point.\n")

logger.info("Press enter to START trigger:")
input()  # Wait for the user to press Enter

def turn_off_console_and_tx():
    if not use_external_power_supply:
        logger.info("Attempting to turn off High Voltage...")
        interface.hvcontroller.turn_hv_off()
        if interface.hvcontroller.get_hv_status():
            logger.error("High Voltage is still on.")
        else:
            logger.info("High Voltage successfully turned off.")

        logger.info("Attempting to turn off 12V...")
        interface.hvcontroller.turn_12v_off()
        if interface.hvcontroller.get_12v_status():
            logger.error("12V is still on.")
        else:
            logger.info("12V successfully turned off.")

shutdown_event = threading.Event()

def input_wrapper():
    if os.name == 'nt':
        while not shutdown_event.is_set():
            if msvcrt.kbhit():  # Check if a key has been pressed
                char = msvcrt.getch()  # Read the key
                if char == b'\r':  # Check if the Enter key was pressed
                    break
            time.sleep(0.1)  # Small delay to prevent high CPU usage
    else:
        while not shutdown_event.is_set():
            # Use select to check if input is available
            ready, _, _ = select.select([], [], [], 0.1)  # Timeout of 0.1 seconds
            if ready:
                sys.stdin.readline() # Read the input
                break

t = threading.Thread(target=log_temperature)
user_input = threading.Thread(target=input_wrapper)

logger.info("Starting Trigger...")
if interface.start_sonication():
    logger.info("Trigger Running...")
    logger.info("Press enter to STOP trigger:")

    # Start logging and user input threads
    t.start()
    user_input.start()

    while (user_input.is_alive() and t.is_alive()):
        time.sleep(0.1)

    if (not user_input.is_alive()):
        logger.info("Logging interrupted by user.")
    else:
        logger.info("Logging interrupted by shutdown event.")
        shutdown_event.set()  # Signal the logging thread to stop
    stop_logging = True

    # Wait for threads to finish
    t.join()
    user_input.join()

    time.sleep(0.5)  # Give the logging thread time to finish
    if interface.stop_sonication():
        logger.info("Trigger stopped successfully.")
    else:
        logger.error("Failed to stop trigger.")
else:
    logger.error("Failed to get trigger setting.")

turn_off_console_and_tx()



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: env
#     language: python
#     name: python3
# ---
from __future__ import annotations

import numpy as np

from openlifu.bf import Pulse, Sequence, apod_methods, focal_patterns
from openlifu.geo import Point
from openlifu.plan import Protocol
from openlifu.plan.param_constraint import ParameterConstraint
from openlifu.sim import SimSetup
from openlifu.xdc import Transducer

# +
f0 = 400e3
pulse = Pulse(frequency=f0, duration=10/f0)
sequence = Sequence(pulse_interval=0.1, pulse_count=9, pulse_train_interval=0, pulse_train_count=1)
focal_pattern = focal_patterns.SinglePoint(target_pressure=1.2e6)
focal_pattern = focal_patterns.Wheel(center=False, spoke_radius=5, num_spokes=3, target_pressure=1.2e6)
apod_method = apod_methods.MaxAngle(30)
sim_setup = SimSetup(x_extent=(-30,30), y_extent=(-30,30), z_extent=(-4,70))
protocol = Protocol(
    id='test_protocol',
    name='Test Protocol',
    pulse=pulse,
    sequence=sequence,
    focal_pattern=focal_pattern,
    apod_method=apod_method,
    sim_setup=sim_setup)

target = Point(position=np.array([0, 0, 50]), units="mm", radius=2)
trans = Transducer.gen_matrix_array(nx=8, ny=8, pitch=4, kerf=0.5, id="m3", name="openlifu", sensitivity=1e6/10)
# -

solution, sim_res, scaled_analysis = protocol.calc_solution(
    target=target,
    transducer=trans,
    simulate=True,
    scale=True)

pc = {"MI":ParameterConstraint('<', 1.8, 1.85), "TIC":ParameterConstraint('<', 2.0), 'global_isppa_Wcm2':ParameterConstraint('within', error_value=(49, 190))}
if scaled_analysis is not None:
    scaled_analysis.to_table(constraints=pc).set_index('Param')[['Value', 'Units', 'Status']]

protocol = Protocol.from_file('../tests/resources/example_db/protocols/example_protocol/example_protocol.json')
solution, sim_res, analysis = protocol.calc_solution(
    target=target,
    transducer=trans,
    simulate=True,
    scale=True)
if analysis is not None:
    analysis.to_table().set_index('Param')[['Value', 'Units', 'Status']]



from __future__ import annotations

import glob
import json
import logging
import os
import shutil
import tempfile
from enum import Enum
from pathlib import Path
from typing import Dict, List

import h5py
import nibabel as nib

from openlifu.nav.photoscan import Photoscan, load_data_from_photoscan
from openlifu.plan import Protocol, Run, Solution
from openlifu.util.json import PYFUSEncoder
from openlifu.util.types import PathLike
from openlifu.util.volume_conversion import (
    convert_dicom_to_nifti,
    is_dicom_file_or_directory,
)
from openlifu.xdc import Transducer, TransducerArray
from openlifu.xdc.util import load_transducer_from_file

from .session import Session
from .subject import Subject
from .user import User

OnConflictOpts = Enum('OnConflictOpts', ['ERROR', 'OVERWRITE', 'SKIP'])


class Database:
    def __init__(self, path: str | None = None):
        if path is None:
            path = Database.get_default_path()
        self.path = os.path.normpath(path)
        self.logger = logging.getLogger(__name__)

    def write_gridweights(self, transducer_id: str, grid_hash: str, grid_weights, on_conflict: OnConflictOpts = OnConflictOpts.ERROR):
        grid_hashes = self.get_gridweight_hashes(transducer_id)
        if grid_hash in grid_hashes:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Grid weights with hash {grid_hash} already exists for transducer {transducer_id}.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting grid weights with hash {grid_hash} for transducer {transducer_id}.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping grid weights with hash {grid_hash} for transducer {transducer_id} as it already exists.")
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")
        gridweight_filename = self.get_gridweights_filename(transducer_id, grid_hash)
        with h5py.File(gridweight_filename, "w") as f:
            f.create_dataset("grid_weights", data=grid_weights)
        self.logger.info(f"Added grid weights with hash {grid_hash} for transducer {transducer_id} to the database.")

    def write_user(self, user: User, on_conflict: OnConflictOpts = OnConflictOpts.ERROR) -> None:
        # Check if the sonication user ID already exists in the database
        user_id = user.id
        user_ids = self.get_user_ids()

        if user_id in user_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"User with ID {user_id} already exists in the database.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting User with ID {user_id} in the database.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping User with ID {user_id} as it already exists in the database.")
                return  # Skip adding the User

        # Save the user to a JSON file
        user_filename = self.get_user_filename(user_id)
        user.to_file(str(user_filename))

        # Update the list of User IDs
        if user_id not in user_ids:
            user_ids.append(user_id)
            self.write_user_ids(user_ids)

        self.logger.info(f"Added User with ID {user_id} to the database.")

    def delete_user(self, user_id: str, on_conflict: OnConflictOpts = OnConflictOpts.ERROR) -> None:
        # Check if the user ID already exists in the database
        user_ids = self.get_user_ids()

        if user_id not in user_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"User ID {user_id} does not exist in the database.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Cannot delete user ID {user_id} as it does not exist in the database.")
                return
            else:
                raise ValueError("Invalid 'on_conflict' option.")

        # Delete the directory of the user
        user_dir = self.get_user_dir(user_id)
        if Path.is_dir(user_dir):
            shutil.rmtree(user_dir)

        if user_id in user_ids:
            user_ids.remove(user_id)
            self.write_user_ids(user_ids)

        self.logger.info(f"Removed Sonication User with ID {user_id} from the database.")

    def write_protocol(self, protocol: Protocol, on_conflict: OnConflictOpts = OnConflictOpts.ERROR):
        # Check if the sonication protocol ID already exists in the database
        protocol_id = protocol.id
        protocol_ids = self.get_protocol_ids()

        if protocol_id in protocol_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Protocol with ID {protocol_id} already exists in the database.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting Protocol with ID {protocol_id} in the database.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping Protocol with ID {protocol_id} as it already exists in the database.")
                return  # Skip adding the Protocol
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

        # Save the sonication protocol to a JSON file
        protocol_filename = self.get_protocol_filename(protocol_id)
        protocol.to_file(protocol_filename)

        # Update the list of Protocol IDs
        if protocol_id not in protocol_ids:
            protocol_ids.append(protocol_id)
            self.write_protocol_ids(protocol_ids)

        self.logger.info(f"Added Sonication Protocol with ID {protocol_id} to the database.")

    def delete_protocol(self, protocol_id: str, on_conflict: OnConflictOpts = OnConflictOpts.ERROR):
        # Check if the sonication protocol ID already exists in the database
        protocol_ids = self.get_protocol_ids()

        if protocol_id not in protocol_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Protocol ID {protocol_id} does not exist in the database.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Cannot delete protocol ID {protocol_id} as it does not exist in the database.")
                return
            else:
                raise ValueError("Invalid 'on_conflict' option.")

        # Delete the directory of the protocol
        protocol_dir = self.get_protocol_dir(protocol_id)
        if Path.is_dir(protocol_dir):
            shutil.rmtree(protocol_dir)

        if protocol_id in protocol_ids:
            protocol_ids.remove(protocol_id)
            self.write_protocol_ids(protocol_ids)

        self.logger.info(f"Removed Sonication Protocol with ID {protocol_id} from the database.")

    def write_session(self, subject:Subject, session:Session, on_conflict=OnConflictOpts.ERROR):
        # Generate session ID
        session_id = session.id

        # Validate the subject ID in the session
        if session.subject_id != subject.id:
            raise ValueError("IDs do not match between the given subject and the subject referenced in the session.")

        # Validate the virtual fit results
        for target_id, list_of_transforms in session.virtual_fit_results.items():
            if target_id not in [target.id for target in session.targets]:
                raise ValueError(
                    f"The virtual_fit_results of session {session.id} references a target {target_id} that is not"
                    " in the session's list of targets."
                )
            if len(list_of_transforms)<1:
                raise ValueError(
                    f"The virtual_fit_results of session {session.id} provides no transforms for target {target_id}."
                )

        # Validate the transducer tracking result's photoscan_id
        if session.transducer_tracking_results:
            for result in session.transducer_tracking_results:
                if result.photoscan_id not in self.get_photoscan_ids(subject.id, session.id):
                    raise ValueError(
                f"Photoscan id {result.photoscan_id} provided in the transducer_tracking_results has not "
                "been associated with this session."
            )

        # Check if the session already exists in the database
        session_ids = self.get_session_ids(subject.id)

        if session_id in session_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Session with ID {session_id} already exists for subject {subject.id}.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting session with ID {session_id} for subject {subject.id}.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping session with ID {session_id} for subject {subject.id} as it already exists.")
                return  # Skip adding the session
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

        # Save the session to a JSON file
        session_filename = self.get_session_filename(subject.id, session_id)
        session.update_modified_time()
        session.to_file(session_filename)

        # Create empty runs.json, solutions.json and photoscans.json files for session if needed
        if not self.get_solutions_filename(subject.id, session_id).exists():
            self.write_solution_ids(session, [])
        if not self.get_runs_filename(subject.id, session_id).exists():
            self.write_run_ids(session.subject_id, session.id, [])
        if not self.get_photocollections_filename(subject.id, session_id).exists():
            self.write_reference_numbers(session.subject_id, session.id, [])
        if not self.get_photoscans_filename(subject.id, session_id).exists():
            self.write_photoscan_ids(session.subject_id, session.id, [])

        # Update the list of session IDs for the subject
        if session_id not in session_ids:
            session_ids.append(session_id)
            self.write_session_ids(subject.id, session_ids)

        self.logger.info(f"Added session with ID {session_id} for subject {subject.id} to the database.")

    def delete_session(self, subject_id: str, session_id: str, on_conflict: OnConflictOpts = OnConflictOpts.ERROR):
        """Delete a session and its associated data from the database.

        Args:
            subject_id: ID of the subject the session belongs to
            session_id: ID of the session to delete
            on_conflict: Behavior when session doesn't exist ('error' or 'skip')
        """
        # Check if the session ID exists in the database for this subject
        session_ids = self.get_session_ids(subject_id)

        if session_id not in session_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Session ID {session_id} does not exist in the database.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Cannot delete session ID {session_id} as it does not exist in the database.")
                return
            else:
                raise ValueError("Invalid 'on_conflict' option.")

        # Delete the session file
        session_filename = self.get_session_filename(subject_id, session_id)
        if session_filename.parent.is_dir():
            shutil.rmtree(session_filename.parent)

        session_ids.remove(session_id)
        self.write_session_ids(subject_id, session_ids)

        self.logger.info(f"Removed session with ID {session_id} from the database.")

    def write_run(self, run:Run, session:Session = None, protocol:Protocol = None, on_conflict=OnConflictOpts.ERROR):
        """Write a run with a snapshot of session and a snapshot of protocol if provided

        Args:
            run: Run to be written
            session (optional): the Session used in the Run, it will have a snapshot written alongside the Run
            protocol (optional): the Protocol used in the Run, it will have a snapshot written alongside the Run

        Returns:
            None: This method does not return a value
        """
        # Check whether the run already exist in the session
        run_ids = self.get_run_ids(session.subject_id, session.id)

        if run.id in run_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Run with ID {run.id} already exists for session {session.id}.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                raise ValueError("Runs are write-once objects and may not be overwritten.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping run with ID {run.id} for session {session.id} as it already exists.")
                return  # Skip adding the session
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

        # Save the run metadata to a JSON file
        run_metadata_filepath = self.get_run_filepath(session.subject_id, session.id, run.id)
        run.to_file(run_metadata_filepath)

        # Update
        run_ids.append(run.id)
        self.write_run_ids(session.subject_id, session.id, run_ids)

        if session:
            # Write snapshot of the session
            session.to_file(run_metadata_filepath.parent / f'{run.id}_session_snapshot.json')

        if protocol:
            # Write snapshot of the protocol
            protocol.to_file(run_metadata_filepath.parent / f'{run.id}_protocol_snapshot.json')

    def write_subject(self, subject, on_conflict=OnConflictOpts.ERROR):
        subject_id = subject.id
        subject_ids = self.get_subject_ids()

        if subject_id in subject_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Subject with ID {subject_id} already exists in the database.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting subject with ID {subject_id} in the database.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping subject with ID {subject_id} as it already exists.")
                return
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

        subject_filename = self.get_subject_filename(subject_id)
        subject.to_file(subject_filename)

        # Create empty sessions.json and volumes.json files for subject if needed
        if not self.get_sessions_filename(subject.id).exists():
            self.write_session_ids(subject_id, session_ids=[])
        if not self.get_volumes_filename(subject.id).exists():
            self.write_volume_ids(subject_id, volume_ids=[])

        if subject_id not in subject_ids:
            subject_ids.append(subject_id)
            self.write_subject_ids(subject_ids)

        self.logger.info(f"Added subject with ID {subject_id} to the database.")

    def write_transducer(
            self,
            transducer,
            registration_surface_model_filepath: PathLike | None = None,
            transducer_body_model_filepath: PathLike | None = None,
            on_conflict: OnConflictOpts=OnConflictOpts.ERROR,
    ) -> None:
        """ Writes a transducer object to database and copies the affiliated transducer data files to the database if provided. When a transducer that is already present in the database is being re-written,
        the associated model data files do not need to be provided if they have previously been added to the database.
        Args:
            transducer: Transducer to be written
            transducer_body_model_filepath (optional): Model file containing a mesh of transducer body mesh. This is a closed surface meant for visualization of the transducer.
            registration_surface_model_filepath (optional): Model file containing an open-surface sub-mesh of the transducer body model. This model is meant to be used for registration during transducer tracking.
        Returns:
            None: This method does not return a value
        """
        transducer_id = transducer.id
        transducer_ids = self.get_transducer_ids()

        if transducer_id in transducer_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Transducer with ID {transducer_id} already exists in the database.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting transducer with ID {transducer_id} in the database.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping transducer with ID {transducer_id} as it already exists.")
                return
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

        transducer_filename = self.get_transducer_filename(transducer_id)
        transducer_parent_dir = transducer_filename.parent
        transducer_parent_dir.mkdir(exist_ok = True)

        # Copy the transducer data files to database.
        # If tranducer data files not provided, check that any files previously
        # associated with the transducer object exist.
        if registration_surface_model_filepath:
            registration_surface_model_filepath = Path(registration_surface_model_filepath)
            if not registration_surface_model_filepath.exists():
                raise FileNotFoundError(f'Registration surface model filepath does not exist: {registration_surface_model_filepath}')
            transducer.registration_surface_filename = registration_surface_model_filepath.name
            shutil.copy(registration_surface_model_filepath, transducer_parent_dir)
        elif transducer.registration_surface_filename:
            if not (transducer_parent_dir/transducer.registration_surface_filename).exists():
                raise ValueError(f"Cannot find registration surface file associated with transducer {transducer.id}.")

        if transducer_body_model_filepath:
            transducer_body_model_filepath = Path(transducer_body_model_filepath)
            if not transducer_body_model_filepath.exists():
                raise FileNotFoundError(f'Transducer body model filepath does not exist: {transducer_body_model_filepath}')
            transducer.transducer_body_filename = transducer_body_model_filepath.name
            shutil.copy(transducer_body_model_filepath, transducer_parent_dir)
        elif transducer.transducer_body_filename:
            if not (transducer_parent_dir/transducer.transducer_body_filename).exists():
                raise ValueError(f"Cannot find transducer body file associated with transducer {transducer.id}.")

        transducer.to_file(transducer_filename)

        if transducer_id not in transducer_ids:
            transducer_ids.append(transducer_id)
            self.write_transducer_ids(transducer_ids)

        self.logger.info(f"Added transducer with ID {transducer_id} to the database.")

    def write_system(self, system, on_conflict=OnConflictOpts.ERROR):
        system_id = system.id
        system_ids = self.get_system_ids()

        if system_id in system_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Ultrasound system with ID {system_id} already exists in the database.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting ultrasound system with ID {system_id} in the database.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping ultrasound system with ID {system_id} as it already exists.")
                return
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

        system_data = system.to_json()
        system_filename = self.get_system_filename(system_id)
        with open(system_filename, "w") as f:
            json.dump(system_data, f)

        if system_id not in system_ids:
            system_ids.append(system_id)
            self.write_system_ids(system_ids)

        self.logger.info(f"Added ultrasound system with ID {system_id} to the database.")

    def write_volume(self, subject_id, volume_id, volume_name, volume_data_filepath, on_conflict=OnConflictOpts.ERROR):
        if not Path(volume_data_filepath).exists():
            raise ValueError(f'Volume data filepath does not exist: {volume_data_filepath}')

        path = Path(volume_data_filepath)
        if path.is_dir() and not is_dicom_file_or_directory(volume_data_filepath):
            raise ValueError(f'Volume data filepath is a directory without DICOM files: {volume_data_filepath}')

        # convert dicom to nifti if needed
        temp_nifti_path = None
        if is_dicom_file_or_directory(volume_data_filepath):
            self.logger.info(f"Detected DICOM input for volume {volume_id}, converting to NIfTI format")
            with tempfile.NamedTemporaryFile(suffix='.nii.gz', delete=False) as temp_file:
                temp_nifti_path = Path(temp_file.name)

            try:
                convert_dicom_to_nifti(volume_data_filepath, temp_nifti_path)
                volume_data_filepath = temp_nifti_path
            except Exception as e:
                if temp_nifti_path.exists():
                    temp_nifti_path.unlink()
                raise RuntimeError(f"Failed to convert DICOM to NIfTI: {e}") from e

        try:
            volume_ids = self.get_volume_ids(subject_id)
            if volume_id in volume_ids:
                if on_conflict == OnConflictOpts.ERROR:
                    raise ValueError(f"Volume with ID {volume_id} already exists for subject {subject_id}.")
                elif on_conflict == OnConflictOpts.OVERWRITE:
                    self.logger.info(f"Overwriting volume with ID {volume_id} for subject {subject_id}.")
                elif on_conflict == OnConflictOpts.SKIP:
                    self.logger.info(f"Skipping volume with ID {volume_id} for subject {subject_id} as it already exists.")
                    return
                else:
                    raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

            volume_metadata_dict = {"id": volume_id, "name": volume_name, "data_filename": Path(volume_data_filepath).name}
            volume_metadata_json = json.dumps(volume_metadata_dict, separators=(',', ':'), cls=PYFUSEncoder)

            volume_metadata_filepath = self.get_volume_metadata_filepath(subject_id, volume_id)
            Path(volume_metadata_filepath).parent.parent.mkdir(exist_ok=True)
            Path(volume_metadata_filepath).parent.mkdir(exist_ok=True)
            with open(volume_metadata_filepath, 'w') as file:
                file.write(volume_metadata_json)
            shutil.copy(Path(volume_data_filepath), Path(volume_metadata_filepath).parent)

            if volume_id not in volume_ids:
                volume_ids.append(volume_id)
                self.write_volume_ids(subject_id, volume_ids)

            self.logger.info(f"Added volume with ID {volume_id} for subject {subject_id} to the database.")
        finally:
            # cleanup temp nifti file
            if temp_nifti_path is not None and temp_nifti_path.exists():
                temp_nifti_path.unlink()

    def write_photocollection(self, subject_id, session_id, reference_number: str, photo_paths: List[PathLike], on_conflict=OnConflictOpts.ERROR):
        """ Writes a photocollection to database and copies the associated
        photos into the database, specified by the subject, session, and
        reference_number of the photocollection."""

        photocollection_dir = Path(self.get_session_dir(subject_id, session_id)) / 'photocollections' / reference_number

        reference_numbers = self.get_photocollection_reference_numbers(subject_id, session_id)
        if reference_number in reference_numbers:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Photocollection with reference number {reference_number} already exists for session {session_id}.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting photocollection with reference number {reference_number} for session {session_id}.")
                if photocollection_dir.exists():
                    shutil.rmtree(photocollection_dir)
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping photocollection with reference number {reference_number} for session {session_id} as it already exists.")
                return
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

        photocollection_dir.mkdir(exist_ok=True)

        # Copy each photo into the photocollection directory
        for photo_path in photo_paths:
            photo_path = Path(photo_path)
            if not photo_path.exists():
                raise FileNotFoundError(f"Photo file does not exist: {photo_path}")
            shutil.copy(photo_path, photocollection_dir)

        if reference_number not in reference_numbers:
            reference_numbers.append(reference_number)
            self.write_reference_numbers(subject_id,session_id, reference_numbers)

        self.logger.info(f"Added photocollection with reference number {reference_number} for session {session_id} to the database.")

    def write_photoscan(self, subject_id, session_id, photoscan: Photoscan, model_data_filepath: str | None = None, texture_data_filepath: str | None = None, mtl_data_filepath: str | None = None, on_conflict=OnConflictOpts.ERROR):
        """ Writes a photoscan object to database and copies the associated data filepaths into the database.
        While the model data file is required, the associated texture and .mtl files are optional and can be provided if present.
        When a photoscan that is already present in the database is being re-written,the associated data files do not need to be specified """

        photoscan_ids = self.get_photoscan_ids(subject_id, session_id)
        if photoscan.id in photoscan_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Photoscan with ID {photoscan.id} already exists for session {session_id}.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting photoscan with ID {photoscan.id} for session {session_id}.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping photoscan with ID {photoscan.id} for session {session_id} as it already exists.")
                return
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

        photoscan_metadata_filepath = Path(self.get_photoscan_metadata_filepath(subject_id, session_id, photoscan.id)) #subject_id/photoscan/photoscan_id/photoscan_id.json
        photoscan_parent_dir = photoscan_metadata_filepath.parent
        photoscan_parent_dir.mkdir(exist_ok=True)

        # Copy photoscan model and texture files to database.
        # If a model and texture data file is not provided, check that there are existing files previously
        # associated with the photoscan object.
        if model_data_filepath:
            model_data_filepath = Path(model_data_filepath)
            if not model_data_filepath.exists():
                raise FileNotFoundError(f'Model data filepath does not exist: {model_data_filepath}')
            photoscan.model_filename = model_data_filepath.name
            shutil.copy(model_data_filepath, photoscan_metadata_filepath.parent)
        elif not photoscan.model_filename or not (photoscan_parent_dir/photoscan.model_filename).exists():
            raise ValueError(f"Cannot find model file associated with photoscan {photoscan.id}.")

        if texture_data_filepath:
            texture_data_filepath = Path(texture_data_filepath)
            if not texture_data_filepath.exists():
                raise FileNotFoundError(f'Texture data filepath does not exist: {texture_data_filepath}')
            photoscan.texture_filename = texture_data_filepath.name
            shutil.copy(texture_data_filepath, photoscan_metadata_filepath.parent)
        elif photoscan.texture_filename and not (photoscan_parent_dir/photoscan.texture_filename).exists():
            raise ValueError(f"Cannot find texture file associated with photoscan {photoscan.id}.")

        # Not necessarily required for a photoscan object
        if mtl_data_filepath:
            mtl_data_filepath = Path(mtl_data_filepath)
            if not mtl_data_filepath.exists():
                raise FileNotFoundError(f'MTL filepath does not exist: {mtl_data_filepath}')
            photoscan.mtl_filename = mtl_data_filepath.name
            shutil.copy(mtl_data_filepath, photoscan_metadata_filepath.parent)
        elif photoscan.mtl_filename and not (photoscan_parent_dir/photoscan.mtl_filename).exists():
            raise ValueError(f"Cannot find photoscan materials file associated with photoscan {photoscan.id}.")

        #Save the photoscan metadata to a JSON file
        photoscan.to_file(photoscan_metadata_filepath)

        if photoscan.id not in photoscan_ids:
            photoscan_ids.append(photoscan.id)
            self.write_photoscan_ids(subject_id,session_id, photoscan_ids)

        self.logger.info(f"Added photoscan with ID {photoscan.id} for session {session_id} to the database.")

    def write_solution(self, session:Session, solution:Solution, on_conflict: OnConflictOpts=OnConflictOpts.ERROR):
        solution_ids = self.get_solution_ids(session.subject_id, session.id)

        if solution.id in solution_ids:
            if on_conflict == OnConflictOpts.ERROR:
                raise ValueError(f"Solution with ID {solution.id} already exists in the database.")
            elif on_conflict == OnConflictOpts.OVERWRITE:
                self.logger.info(f"Overwriting solution with ID {solution.id} in the database.")
            elif on_conflict == OnConflictOpts.SKIP:
                self.logger.info(f"Skipping solution with ID {solution.id} as it already exists.")
                return
            else:
                raise ValueError("Invalid 'on_conflict' option. Use 'error', 'overwrite', or 'skip'.")

        solution_json_filepath = self.get_solution_filepath(session.subject_id, session.id, solution.id)
        solution.to_files(solution_json_filepath)

        if solution.id not in solution_ids:
            solution_ids.append(solution.id)
            self.write_solution_ids(session, solution_ids)

        self.logger.info(f"Wrote solution with ID {solution.id} to the database.")

    def choose_session(self, subject, options=None):
        # Implement the logic to choose a session
        raise NotImplementedError("Method not yet implemented")

    def choose_subject(self, options=None):
        # Implement the logic to choose a subject
        raise NotImplementedError("Method not yet implemented")

    def get_gridweight_hashes(self, transducer_id):
        transducer_dir = Path(self.path)/'transducers'/transducer_id
        gridfiles = glob.glob(Path(transducer_dir)/f'{transducer_id}_gridweights_*.h5')
        return [os.path.splitext(os.path.basename(f))[0].split('_')[-1] for f in gridfiles]

    def get_session_table(self, subject_id, options=None):
        # Implement the logic to get session table
        raise NotImplementedError("Method not yet implemented")

    def get_subject_table(self, options=None):
        # Implement the logic to get subject table
        raise NotImplementedError("Method not yet implemented")

    def get_connected_systems(self):
        connected_system_filename = self.get_connected_system_filename()

        if os.path.isfile(connected_system_filename):
            with open(connected_system_filename) as file:
                connected_systems = file.read().strip().split(',')
            self.logger.info("Connected systems: %s", connected_systems)
            return connected_systems
        else:
            self.logger.warning("Connected systems file not found.")
            return []

    def get_connected_transducer(self, options=None):
        connected_transducer_filename = self.get_connected_transducer_filename()

        if os.path.isfile(connected_transducer_filename):
            with open(connected_transducer_filename) as file:
                connected_transducer_id = file.read().strip()
            self.logger.info("Connected transducer: %s", connected_transducer_id)
            return connected_transducer_id
        else:
            self.logger.warning("Connected transducer file not found.")
            return None

    def get_user_ids(self) -> List[str]:
        users_filename = self.get_users_filename()

        if os.path.isfile(users_filename):
            with open(users_filename) as file:
                user_data = json.load(file)
                user_ids = user_data.get("user_ids", [])
            self.logger.info("User IDs: %s", user_ids)
            return user_ids
        else:
            self.logger.warning("Users file not found.")
            return []

    def get_protocol_ids(self):
        protocols_filename = self.get_protocols_filename()

        if os.path.isfile(protocols_filename):
            with open(protocols_filename) as file:
                protocol_data = json.load(file)
                protocol_ids = protocol_data.get("protocol_ids", [])
            self.logger.info("Protocol IDs: %s", protocol_ids)
            return protocol_ids
        else:
            self.logger.warning("Protocols file not found.")
            return []

    def get_session_ids(self, subject_id):
        sessions_filename = self.get_sessions_filename(subject_id)

        if os.path.isfile(sessions_filename):
            with open(sessions_filename) as file:
                session_data = json.load(file)
                session_ids = session_data.get("session_ids", [])
            self.logger.info("Session IDs for subject %s: %s", subject_id, session_ids)
            return session_ids
        else:
            self.logger.info("Sessions file not found for subject %s.", subject_id)
            return []

    def get_run_ids(self, subject_id, session_id):
        runs_filename = self.get_runs_filename(subject_id, session_id)

        if os.path.isfile(runs_filename):
            with open(runs_filename) as file:
                run_data = json.load(file)
                run_ids = run_data.get("run_ids", [])
            self.logger.info("Run IDs for session %s: %s", session_id, run_ids)
            return run_ids
        else:
            self.logger.warning("Runs file not found for session %s.", session_id)
            return []

    def get_volume_ids(self, subject_id):
        volumes_filename = self.get_volumes_filename(subject_id)
        if os.path.isfile(volumes_filename):
            with open(volumes_filename) as file:
                volume_data = json.load(file)
                volume_ids = volume_data.get("volume_ids", [])
            self.logger.info("Volume IDs for subject %s: %s", subject_id, volume_ids)
            return volume_ids
        else:
            self.logger.info("Volumes file not found for subject %s.", subject_id)
            return []

    def get_solution_ids(self, subject_id:str, session_id:str) -> List[str]:
        """Get a list of IDs of the solutions associated with the given session"""
        solutions_filename = self.get_solutions_filename(subject_id, session_id)

        if not (solutions_filename.exists() and solutions_filename.is_file()):
            self.logger.warning("Solutions file not found for subject %s, session %s.", subject_id, session_id)
            return []

        return json.loads(solutions_filename.read_text())["solution_ids"]

    def get_photocollection_reference_numbers(self, subject_id: str, session_id: str) -> List[str]:
        """Get a list of reference numbers of the photocollections associated with the given session"""
        photocollection_filename = self.get_photocollections_filename(subject_id, session_id)

        if not (photocollection_filename.exists() and photocollection_filename.is_file()):
            self.logger.warning("Photocollection file not found for subject %s, session %s.", subject_id, session_id)
            return []

        return json.loads(photocollection_filename.read_text())["reference_numbers"]

    def get_photoscan_ids(self, subject_id: str, session_id: str) -> List[str]:
        """Get a list of IDs of the photoscans associated with the given session"""
        photoscan_filename = self.get_photoscans_filename(subject_id, session_id)

        if not (photoscan_filename.exists() and photoscan_filename.is_file()):
            self.logger.warning("Photoscan file not found for subject %s, session %s.", subject_id, session_id)
            return []

        return json.loads(photoscan_filename.read_text())["photoscan_ids"]

    def get_subject_ids(self):
        subjects_filename = self.get_subjects_filename()

        if os.path.isfile(subjects_filename):
            with open(subjects_filename) as file:
                subject_data = json.load(file)
                subject_ids = subject_data.get("subject_ids", [])
            self.logger.info("Subject IDs: %s", subject_ids)
            return subject_ids
        else:
            self.logger.warning("Subjects file not found.")
            return []

    def get_system_ids(self):
        systems_filename = self.get_systems_filename()

        if os.path.isfile(systems_filename):
            with open(systems_filename) as file:
                system_data = json.load(file)
                system_ids = system_data.get("system_ids", [])
            self.logger.info("System IDs: %s", system_ids)
            return system_ids
        else:
            self.logger.warning("Systems file not found.")
            return []

    def get_system_info(self, sys_id):
        raise NotImplementedError("UltrasoundSystem is not yet implemented")
        system_filename = self.get_system_filename(sys_id)

        if os.path.isfile(system_filename):
            with open(system_filename) as file:
                system_data = json.load(file)
                system_info = UltrasoundSystem.from_dict(system_data)
            self.logger.info("System info for system %s: %s", sys_id, system_info)
            return system_info
        else:
            self.logger.warning("System info file not found for system %s.", sys_id)
            return None

    def get_transducer_ids(self):
        transducers_filename = self.get_transducers_filename()

        if os.path.isfile(transducers_filename):
            with open(transducers_filename) as file:
                transducer_data = json.load(file)
                transducer_ids = transducer_data.get("transducer_ids", [])
            if not isinstance(transducer_ids, list):
                transducer_ids = [transducer_ids]
            self.logger.info("Transducer IDs: %s", transducer_ids)
            return transducer_ids
        else:
            self.logger.warning("Transducers file not found.")
            return []

    def load_gridweights(self, transducer_id, grid_hash):
        gridweight_filename = self.get_gridweights_filename(transducer_id, grid_hash)
        with h5py.File(gridweight_filename, "r") as f:
            grid_weights = f["grid_weights"][:]
        return grid_weights

    def load_subject(self, subject_id, options=None):
        subject_filename = self.get_subject_filename(subject_id)
        subject = Subject.from_file(subject_filename)
        self.logger.info(f"Loaded subject {subject_id}")
        return subject

    def get_volume_info(self, subject_id, volume_id):
        volume_metadata_filepath = self.get_volume_metadata_filepath(subject_id, volume_id)
        with open(volume_metadata_filepath) as f:
            volume = json.load(f)
            return {"id": volume["id"],\
                    "name": volume["name"],\
                    "data_abspath": Path(volume_metadata_filepath).parent/volume["data_filename"]}

    def get_photocollection_absolute_filepaths(self, subject_id: str, session_id: str, reference_number: str) -> List[Path]:
        """
        get the absolute filepaths of all photos in a specific photocollection.

        Args:
            subject_id (str): The subject ID.
            session_id (str): The session ID.
            reference_number (str): The reference number of the photocollection.

        Returns:
            List[Path]: List of absolute file paths to the photos in the photocollection.
        """
        photocollection_dir = (
            Path(self.get_session_dir(subject_id, session_id)) / 'photocollections' / reference_number
        )

        if not photocollection_dir.exists() or not photocollection_dir.is_dir():
            self.logger.warning(
                f"Photocollection directory not found for subject {subject_id}, "
                f"session {session_id}, photocollection {reference_number}."
            )
            return []

        return sorted(photocollection_dir.glob("*"))

    def get_photoscan_absolute_filepaths_info(self, subject_id, session_id, photoscan_id):
        """Returns the photoscan information with absolute paths to any data"""
        photoscan_metadata_filepath = self.get_photoscan_metadata_filepath(subject_id, session_id, photoscan_id)
        photoscan_metadata_directory = Path(photoscan_metadata_filepath).parent
        with open(photoscan_metadata_filepath) as f:
            photoscan = json.load(f)
            photoscan_dict = {"id": photoscan["id"],\
                    "name": photoscan["name"],\
                    "model_abspath": photoscan_metadata_directory/photoscan["model_filename"],
                    "texture_abspath": photoscan_metadata_directory/photoscan["texture_filename"],
                    "photoscan_approved": photoscan["photoscan_approved"]}
            if "mtl_filename" in photoscan:
                photoscan_dict["mtl_abspath"] = Path(photoscan_metadata_filepath).parent/photoscan["mtl_filename"]
        return photoscan_dict

    def load_photoscan(self, subject_id, session_id, photoscan_id, load_data = False):
        """Returns a photoscan object and optionally, also returns the loaded model and texture
        data as Tuple[vtkPolyData, vtkImageData | None ] if load_data = True."""

        photoscan_metadata_filepath = self.get_photoscan_metadata_filepath(subject_id, session_id, photoscan_id)
        if not photoscan_metadata_filepath.exists() or not photoscan_metadata_filepath.is_file():
            raise FileNotFoundError(f"Photoscan file not found for photoscan {photoscan_id}, session {session_id}")

        photoscan = Photoscan.from_file(photoscan_metadata_filepath)

        if load_data:
            (model_data, texture_data) = load_data_from_photoscan(photoscan, Path(photoscan_metadata_filepath.parent))
            return photoscan, (model_data, texture_data)
        return photoscan

    def get_transducer_absolute_filepaths(self, transducer_id:str) -> Dict[str,str | None]:
        """ Returns the absolute filepaths to the model data files i.e. transducer body and registration surface
        model files affiliated with the transducer, with ID `transducer_id`. Unlike `load_transducer`, which
        specifies the relative paths to the model datafiles along with other transducer attributes, this function
        only returns the absolute filepaths to  the datafiles based on the Database directory.

        Args:
            transducer_id: Transducer ID

        Returns:
            dict: A dictionary containing the absolute filepaths to the affiliated transducer data files with the following possible keys:
                - "id" (str): transducer ID
                - "name" (str): transducer name
                - "registration_surface_abspath" (str or None): absolute path to the transducer registration surface (open-surface mesh
                  used for transducer tracking registration). This key is only included if there *is* an affiliated registration surface.
                  None if no registration surface is available.
                - "transducer_body_abspath" (str or None): absolute path to the transducer body model (closed-surface mesh for visualizing the
                  transducer). This key is only included if there *is* an affiliated body model.
                  None if no transducer body is available.
        """
        transducer_metadata_filepath = self.get_transducer_filename(transducer_id)
        transducer = self.load_transducer(transducer_id, convert_array=True)

        transducer_filepaths_dict = {
            "id": transducer.id,
            "name": transducer.name,
            "registration_surface_abspath" : None,
            "transducer_body_abspath" : None,
        }

        if transducer.registration_surface_filename is not None:
            transducer_filepaths_dict["registration_surface_abspath"] = str(
                Path(transducer_metadata_filepath).parent/(transducer.registration_surface_filename)
            )
        if transducer.transducer_body_filename is not None:
            transducer_filepaths_dict["transducer_body_abspath"] = str(
                Path(transducer_metadata_filepath).parent/(transducer.transducer_body_filename)
            )
        return transducer_filepaths_dict

    def load_standoff(self, transducer_id, standoff_id="standoff"):
        raise NotImplementedError("Standoff is not yet implemented")
        standoff_filename = self.get_standoff_filename(transducer_id, standoff_id)
        standoff = Standoff.from_file(standoff_filename)
        return standoff

    def load_system(self, sys_id=None):
        raise NotImplementedError("UltrasoundSystem is not yet implemented")
        sys_id = sys_id or self.get_connected_systems()
        sys_filename = self.get_system_filename(sys_id)
        sys = UltrasoundSystem.from_file(sys_filename)
        return sys

    def load_transducer(self, transducer_id, convert_array:bool = True) -> Transducer|TransducerArray:
        """Given a transducer_id, reads the corresponding transducer file from database and returns a transducer object.
        Note: the transducer object includes the relative path to the affiliated transducer model data. `get_transducer_absolute_filepaths`, should
        be used to obtain the absolute data filepaths based on the Database directory path.
        Args:
            transducer_id: Transducer ID
            convert_array: When enabled, if a TransducerArray is encountered then it is converted to a Transducer.
        Returns:
            Corresponding Transducer object
        """
        return load_transducer_from_file(
            transducer_filepath=self.get_transducer_filename(transducer_id),
            convert_array=convert_array
        )

    def load_transducer_standoff(self, trans, coords, options=None):
        raise NotImplementedError("Standoff is not yet implemented")
        options = options or {}
        standoff_filename = self.get_standoff_filename(trans.id, "standoff_anchors")
        standoff = Standoff.from_file(standoff_filename)
        # Implement the logic to generate binary mask using standoff and coordinates
        mask = generate_standoff_mask(standoff, coords, options)
        return mask

    def load_protocol(self, protocol_id):
        protocol_filename = self.get_protocol_filename(protocol_id)
        if os.path.isfile(protocol_filename):
            protocol = Protocol.from_file(protocol_filename)
            self.logger.info(f"Loaded Protocol {protocol_id}")
            return protocol
        else:
            self.logger.error(f"Protocol file not found for ID: {protocol_id}")
            raise FileNotFoundError(f"Protocol file not found for ID: {protocol_id}")

    def load_all_protocols(self):
        protocols_filename = self.get_protocols_filename()
        if os.path.isfile(protocols_filename):
            with open(protocols_filename) as file:
                data = json.load(file)
                protocol_ids = data.get('protocol_ids', [])
            protocols = []
            for protocol_id in protocol_ids:
                protocol = self.load_protocol(protocol_id)
                protocols.append(protocol)
            return protocols
        else:
            self.logger.error("Protocols file not found.")
            raise FileNotFoundError("Protocols file not found.")

    def load_user(self, user_id) -> User:
        user_filename = self.get_user_filename(user_id)
        if os.path.isfile(user_filename):
            user = User.from_file(user_filename)
            self.logger.info(f"Loaded User {user_id}")
            return user
        else:
            self.logger.error(f"User file not found for ID: {user_id}")
            raise FileNotFoundError(f"User file not found for ID: {user_id}")

    def load_all_users(self) -> List[User]:
        users_filename = self.get_users_filename()
        if os.path.isfile(users_filename):
            with open(users_filename) as file:
                data = json.load(file)
                user_ids = data.get('user_ids', [])
            users = []
            for user_id in user_ids:
                user = self.load_user(user_id)
                users.append(user)
            return users
        else:
            self.logger.error("Users file not found.")
            raise FileNotFoundError("Users file not found.")

    def load_session(self, subject, session_id, options=None):
        if options is None:
            options = {}
        session_filename = self.get_session_filename(subject.id, session_id)
        if os.path.isfile(session_filename):
            session = Session.from_file(session_filename)
            self.logger.info(f"Loaded session {session_id} for subject {subject.id}")
            return session
        else:
            self.logger.error(f"Session file not found for ID: {session_id}")
            raise FileNotFoundError(f"Session file not found for ID: {session_id}")

    def load_session_info(self, subject_id, session_id):
        session_filename = self.get_session_filename(subject_id, session_id)
        if os.path.isfile(session_filename):
            with open(session_filename) as file:
                session_data = json.load(file)
            self.logger.info(f"Loaded session info for session {session_id} of subject {subject_id}")
            return session_data
        else:
            self.logger.error(f"Session file not found for ID: {session_id}")
            raise FileNotFoundError(f"Session file not found for ID: {session_id}")

    def load_session_snapshot(self, subject_id, session_id, run_id):
        path_to_run = self.get_run_dir(subject_id, session_id, run_id)
        return Session.from_file(path_to_run / f'{run_id}_session_snapshot.json')

    def load_protocol_snapshot(self, subject_id, session_id, run_id):
        path_to_run = self.get_run_dir(subject_id, session_id, run_id)
        return Protocol.from_file(path_to_run / f'{run_id}_protocol_snapshot.json')

    def load_solution(self, session:Session, solution_id:str) -> Solution:
        """Load the Solution of the given ID that is associated with the given Session"""
        solution_json_filepath = self.get_solution_filepath(session.subject_id, session.id, solution_id)

        if not solution_json_filepath.exists() or not solution_json_filepath.is_file():
            self.logger.error(f"Solution file not found for solution {solution_id}, session {session.id}")
            raise FileNotFoundError(f"Solution file not found for solution {solution_id}, session {session.id}")

        solution = Solution.from_files(solution_json_filepath)
        self.logger.info(f"Loaded solution {solution_id}")
        return solution

    def load_volume(self, subject, volume_id):
        """Load the volume with the given ID for the specified subject."""
        volume_metadata_filepath = self.get_volume_metadata_filepath(subject.id, volume_id)
        if not volume_metadata_filepath.exists() or not volume_metadata_filepath.is_file():
            self.logger.error(f"Volume metadata file not found for volume {volume_id}, subject {subject.id}")
            raise FileNotFoundError(f"Volume metadata file not found for volume {volume_id}, subject {subject.id}")
        with open(volume_metadata_filepath) as f:
            volume_metadata = json.load(f)
            volume_data_filepath = Path(volume_metadata_filepath).parent / volume_metadata["data_filename"]
        if not volume_data_filepath.exists() or not volume_data_filepath.is_file():
            self.logger.error(f"Volume data file not found for volume {volume_id}, subject {subject.id}")
            raise FileNotFoundError(f"Volume data file not found for volume {volume_id}, subject {subject.id}")
        # Load the volume data using nibabel
        volume_data = nib.load(volume_data_filepath)
        self.logger.info(f"Loaded volume {volume_id} for subject {subject.id}")
        return volume_data

    def set_connected_transducer(self, trans, options=None):
        trans_id = trans.id
        transducer_ids = self.get_transducer_ids()
        if trans_id not in transducer_ids:
            if not options or not options.add_if_missing:
                self.logger.error(f"Invalid Transducer ID {trans_id}. Valid IDs are {', '.join(transducer_ids)}")
                raise ValueError(f"Invalid Transducer ID {trans_id}")
            else:
                self.write_transducer(trans)
        filename = self.get_connected_transducer_filename()
        with open(filename, 'w') as f:
            f.write(trans_id)

    def get_connected_system_filename(self):
        return Path(self.path) / "systems" / "connected_system.txt"

    def get_connected_transducer_filename(self):
        return Path(self.path) / 'transducers' / 'connected_transducer.txt'

    def get_gridweights_filename(self, transducer_id, grid_hash):
        return Path(self.path) / 'transducers' / transducer_id / f'{transducer_id}_gridweights_{grid_hash}.h5'

    def get_protocols_filename(self):
        return Path(self.path) / 'protocols' / 'protocols.json'

    def get_protocol_dir(self, protocol_id):
        return Path(self.path) / 'protocols' / protocol_id

    def get_protocol_filename(self, protocol_id):
        return self.get_protocol_dir(protocol_id) / f'{protocol_id}.json'

    def get_users_filename(self) -> Path:
        return Path(self.path) / 'users' / 'users.json'

    def get_user_dir(self, user_id) -> Path:
        return Path(self.path) / 'users' / user_id

    def get_user_filename(self, user_id) -> Path:
        return self.get_user_dir(user_id) / f'{user_id}.json'

    def get_session_dir(self, subject_id, session_id):
        return Path(self.get_subject_dir(subject_id)) / 'sessions' / session_id

    def get_session_filename(self, subject_id, session_id):
        return Path(self.get_session_dir(subject_id, session_id)) / f'{session_id}.json'

    def get_sessions_filename(self, subject_id) -> Path:
        return Path(self.get_subject_dir(subject_id)) / 'sessions' / 'sessions.json'

    def get_runs_filename(self, subject_id, session_id):
        return Path(self.get_subject_dir(subject_id)) / 'sessions' / f'{session_id}' / 'runs' / 'runs.json'

    def get_volumes_filename(self, subject_id):
        return Path(self.get_subject_dir(subject_id)) / 'volumes' / 'volumes.json'

    def get_solution_filepath(self, subject_id, session_id, solution_id) -> Path:
        """Get the solution json file for the solution with the given ID"""
        session_dir = self.get_session_dir(subject_id, session_id)
        return Path(session_dir) / 'solutions' / solution_id / f"{solution_id}.json"

    def get_solutions_filename(self, subject_id, session_id) -> Path:
        """Get the path to the overall solutions json file for the requested session"""
        session_dir = self.get_session_dir(subject_id, session_id)
        return Path(session_dir) / 'solutions' / 'solutions.json'

    def get_photocollections_filename(self, subject_id, session_id) -> Path:
        """Get the path to the overall photocollections json file for the requested session"""
        session_dir = self.get_session_dir(subject_id, session_id)
        return Path(session_dir) / 'photocollections' / 'photocollections.json'

    def get_photoscans_filename(self, subject_id, session_id) -> Path:
        """Get the path to the overall photoscans json file for the requested session"""
        session_dir = self.get_session_dir(subject_id, session_id)
        return Path(session_dir) / 'photoscans' / 'photoscans.json'

    def get_standoff_filename(self, transducer_id, standoff_id='standoff'):
        return Path(self.path) / 'transducers' / transducer_id / f'{standoff_id}.json'

    def get_subject_dir(self, subject_id):
        return Path(self.path) / 'subjects' / subject_id

    def get_subject_filename(self, subject_id):
        return Path(self.get_subject_dir(subject_id)) / f'{subject_id}.json'

    def get_subjects_filename(self):
        return Path(self.path) / 'subjects' / 'subjects.json'

    def get_systems_filename(self):
        return Path(self.path) / 'systems' / 'systems.json'

    def get_system_filename(self, system_id):
        return Path(self.path) / 'systems' / system_id / f'{system_id}.json'

    def get_transducer_filename(self, transducer_id):
        return Path(self.path) / 'transducers' / transducer_id / f'{transducer_id}.json'

    def get_transducers_filename(self):
        return Path(self.path) / 'transducers' / 'transducers.json'

    def get_volume_dir(self, subject_id, volume_id):
        return Path(self.get_subject_dir(subject_id)) / 'volumes' / volume_id

    def get_volume_metadata_filepath(self, subject_id, volume_id):
        return Path(self.get_volume_dir(subject_id, volume_id)) / f'{volume_id}.json'

    def get_photoscan_metadata_filepath(self, subject_id, session_id, photoscan_id):
        return Path(self.get_session_dir(subject_id, session_id)) / 'photoscans' / photoscan_id / f'{photoscan_id}.json'

    def get_run_dir(self, subject_id, session_id, run_id):
        run_dir = self.get_session_dir(subject_id, session_id) / 'runs' / f'{run_id}'
        return run_dir

    def get_run_filepath(self, subject_id, session_id, run_id):
        return Path(self.get_run_dir(subject_id, session_id, run_id)) / f'{run_id}.json'

    def write_protocol_ids(self, protocol_ids):
        protocol_data = {'protocol_ids': protocol_ids}
        protocols_filename = self.get_protocols_filename()
        with open(protocols_filename, 'w') as f:
            json.dump(protocol_data, f)

    def write_user_ids(self, user_ids: List[str]) -> None:
        user_data = {'user_ids': user_ids}
        users_filename = self.get_users_filename()
        with open(users_filename, 'w') as f:
            json.dump(user_data, f)

    def write_session_ids(self, subject_id, session_ids):
        session_data = {'session_ids': session_ids}
        sessions_filename = self.get_sessions_filename(subject_id)
        Path(sessions_filename).parent.mkdir(exist_ok=True) #sessions directory
        if not sessions_filename.is_file():
            self.logger.info(f"Adding sessions.json file for subject {subject_id} to the database.")
        with open(sessions_filename, 'w') as f:
            json.dump(session_data, f)

    def write_volume_ids(self, subject_id, volume_ids):
        volume_data = {'volume_ids': volume_ids}
        volumes_filename = self.get_volumes_filename(subject_id)
        Path(volumes_filename).parent.mkdir(exist_ok=True) #volumes directory
        if not volumes_filename.is_file():
            self.logger.info(f"Adding volumes.json file with volume_id {volume_ids} for subject {subject_id} to the database.")
        with open(volumes_filename, 'w') as f:
            json.dump(volume_data, f)

    def write_run_ids(self, subject_id, session_id, run_ids):
        run_data = {'run_ids': run_ids}
        runs_filename = self.get_runs_filename(subject_id, session_id)
        Path(runs_filename).parent.mkdir(exist_ok=True)
        if not runs_filename.is_file():
            self.logger.info(f"Adding runs.json file for session {session_id} of subject {subject_id}.")
        with open(runs_filename, 'w') as f:
            json.dump(run_data, f)

    def write_subject_ids(self, subject_ids):
        subject_data = {'subject_ids': subject_ids}
        subjects_filename = self.get_subjects_filename()
        with open(subjects_filename, 'w') as f:
            json.dump(subject_data, f)

    def write_transducer_ids(self, transducer_ids):
        transducers_data = {'transducer_ids': transducer_ids}
        transducers_filename = self.get_transducers_filename()
        with open(transducers_filename, 'w') as f:
            json.dump(transducers_data, f)

    def write_reference_numbers(self, subject_id, session_id, reference_numbers: List[str]):
        photocollection_data = {'reference_numbers': reference_numbers}
        photocollection_filename = self.get_photocollections_filename(subject_id, session_id)
        photocollection_filename.parent.mkdir(exist_ok = True) # Make a photocollection directory in case it does not exist
        with open(photocollection_filename, 'w') as f:
            json.dump(photocollection_data,f)

    def write_photoscan_ids(self, subject_id, session_id, photoscan_ids: List[str]):
        photoscan_data = {'photoscan_ids': photoscan_ids}
        photoscan_filename = self.get_photoscans_filename(subject_id, session_id)
        photoscan_filename.parent.mkdir(exist_ok = True) # Make a photoscan directory in case it does not exist
        with open(photoscan_filename, 'w') as f:
            json.dump(photoscan_data,f)

    def write_solution_ids(self, session:Session, solution_ids:List[str]):
        """Write to the list of overall solution IDs"""
        solutions_data = {'solution_ids': solution_ids}
        solutions_filepath = self.get_solutions_filename(session.subject_id, session.id)
        solutions_filepath.parent.mkdir(exist_ok=True) # Make solutions directory in case it does not exist
        solutions_filepath.write_text(json.dumps(solutions_data))

    @staticmethod
    def get_default_user_dir():
        """
        Get the default user directory for the database

        :returns: Default user directory
        """
        return os.path.expanduser("~")

    @staticmethod
    def get_default_path():
        """
        Get the default path for the database

        :returns: Default path for the database
        """
        return Path(Database.get_default_user_dir()) / "Documents" / "db"

    @staticmethod
    def initialize_empty_database(database_filepath : PathLike) -> Database:
        """
        Initializes an empty database at the given database_filepath
        """
        database_filepath = Path(database_filepath)
        subdirs = ["protocols", "users", "subjects", "transducers", "systems"]
        for subdir in subdirs:
            (database_filepath / subdir).mkdir(parents=True, exist_ok=True)

        new_db = Database(str(database_filepath))

        new_db.write_protocol_ids([])
        new_db.write_user_ids([])
        new_db.write_subject_ids([])
        new_db.write_transducer_ids([])
        return new_db



from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, Literal, Tuple

import pandas as pd

from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.util.dict_conversion import DictMixin

PARAM_STATUS_SYMBOLS = {
    "ok": "✅",
    "warning": "❗",
    "error": "❌"
}

@dataclass
class ParameterConstraint(DictMixin):
    operator: Annotated[Literal['<', '<=', '>', '>=', 'within', 'inside', 'outside', 'outside_inclusive'], OpenLIFUFieldData("Constraint operator", "Constraint operator used to evaluate parameter values")]
    """Constraint operator used to evaluate parameter values"""

    warning_value: Annotated[float | int | Tuple[float | int, float | int] | None, OpenLIFUFieldData("Warning value", "Threshold or range that triggers a warning")] = None
    """Threshold or range that triggers a warning"""

    error_value: Annotated[float | int | Tuple[float | int, float | int] | None, OpenLIFUFieldData("Error value", "Threshold or range that triggers an error")] = None
    """Threshold or range that triggers an error"""

    def __post_init__(self):
        if self.warning_value is None and self.error_value is None:
            raise ValueError("At least one of warning_value or error_value must be set")
        if self.operator in ['within', 'inside', 'outside', 'outside_inclusive']:
            if self.warning_value and (not isinstance(self.warning_value, tuple) or len(self.warning_value) != 2 or self.warning_value[0] >= self.warning_value[1]):
                raise ValueError("Warning value must be a sorted tuple of two numbers")
            if self.error_value and (not isinstance(self.error_value, tuple) or len(self.error_value) != 2 or self.error_value[0] >= self.error_value[1]):
                raise ValueError("Error value must be a sorted tuple of two numbers")
        elif self.operator in ['<', '<=', '>', '>=']:
            if self.warning_value is not None and not isinstance(self.warning_value, int | float):
                raise ValueError("Warning value must be a single value")
            if self.error_value is not None and not isinstance(self.error_value, int | float):
                raise ValueError("Error value must be a single value")

    @staticmethod
    def compare(value, operator, threshold) -> bool:
        if operator == '<':
            return value < threshold
        elif operator == '<=':
            return value <= threshold
        elif operator == '>':
            return value > threshold
        elif operator == '>=':
            return value >= threshold
        elif operator == 'within':
            return threshold[0] < value < threshold[1]
        elif operator == 'inside':
            return threshold[0] <= value <= threshold[1]
        elif operator == 'outside':
            return value < threshold[0] or value > threshold[1]
        elif operator == 'outside_inclusive':
            return value <= threshold[0] or value >= threshold[1]
        else:
            raise ValueError(f"Unsupported operator: {operator}")

    def is_warning(self, value: float | int) -> bool:
        if self.warning_value is not None:
            return not self.compare(value, self.operator, self.warning_value)
        return False

    def is_error(self, value: float | int) -> bool:
        if self.error_value is not None:
            return not self.compare(value, self.operator, self.error_value)
        return False

    def get_status(self, value: float) -> str:
        if self.is_error(value):
            return "error"
        elif self.is_warning(value):
            return "warning"
        else:
            return "ok"

    def get_status_symbol(self, value: float) -> str:
        return PARAM_STATUS_SYMBOLS[self.get_status(value)]

    def to_table(self) -> pd.DataFrame:
        """Convert the parameter constraint to a table format."""
        records = []
        if self.operator in ['<', '<=', '>', '>='] or self.operator in ['within', 'inside', 'outside', 'outside_inclusive']:
            value_template = f"value {self.operator} {{value}}"
        else:
            raise ValueError(f"Unsupported operator: {self.operator}")
        if self.warning_value is not None:
            records.append({"Name": "Warn if not", "Value": value_template.format(value=self.warning_value), "Unit": ""})
        if self.error_value is not None:
            records.append({"Name": "Error if not", "Value": value_template.format(value=self.error_value), "Unit": ""})
        return pd.DataFrame.from_records(records)



from __future__ import annotations

from pathlib import Path

import pytest

from openlifu import User


@pytest.fixture()
def example_user() -> User:
    return User.from_file(Path(__file__).parent/'resources/example_db/users/example_user/example_user.json')

@pytest.mark.parametrize("compact_representation", [True, False])
def test_serialize_deserialize_user(example_user : User, compact_representation: bool):
    assert example_user.from_json(example_user.to_json(compact_representation)) == example_user

def test_default_user():
    """Ensure it is possible to construct a default user"""
    User()



from __future__ import annotations

import asyncio
import logging
from enum import Enum
from typing import Dict, List

import numpy as np
import pandas as pd

import openlifu
from openlifu.io.LIFUHVController import HVController
from openlifu.io.LIFUSignal import LIFUSignal
from openlifu.io.LIFUTXDevice import TriggerModeOpts, TxDevice
from openlifu.io.LIFUUart import LIFUUart
from openlifu.plan.solution import Solution

REF_MAX_SEQUENCE_TIMES = {
    "default": [2*60, 5*60, 10*60],    # users to use default values
    "stress_test": [5*60, 9*60, 15*60] # QA to use stress test values
}

REF_MAX_DUTY_CYCLES = [0.05, 0.1, 0.2, 0.3, 0.4, 0.5]

MAX_VOLTAGE_BY_DUTY_CYCLE_AND_SEQUENCE_TIME = {
    "evt2": [
        [45, 45, 45], # 0.05
        [40, 40, 40], # 0.1
        [40, 40, 35], # 0.2
        [40, 35, 30], # 0.3
        [35, 30, 25], # 0.4
        [30, 25, 20] # 0.5
    ],
    "evt0": [
        [65, 65, 65], # 0.05
        [65, 65, 50], # 0.1
        [50, 40, 35], # 0.2
        [45, 35, 30], # 0.3
        [35, 30, 25], # 0.4
        [30, 25, 20] # 0.5
    ],
}

class LIFUInterfaceStatus(Enum):
    STATUS_COMMS_ERROR = -1
    STATUS_SYS_OFF = 0
    STATUS_SYS_POWERUP = 1
    STATUS_SYS_ON = 2
    STATUS_PROGRAMMING = 3
    STATUS_READY = 4
    STATUS_NOT_READY = 5
    STATUS_RUNNING = 6
    STATUS_FINISHED = 7
    STATUS_ERROR = 8

logger = logging.getLogger(__name__)

class LIFUInterface:
    signal_connect: LIFUSignal = LIFUSignal()
    signal_disconnect: LIFUSignal = LIFUSignal()
    signal_data_received: LIFUSignal = LIFUSignal()
    hvcontroller: HVController = None
    txdevice: TxDevice = None

    def __init__(self,
                 vid: int = 0x0483,
                 tx_pid: int = 0x57AF,
                 con_pid: int = 0x57A0,
                 baudrate: int = 921600,
                 timeout: int = 10,
                 TX_test_mode: bool = False,
                 HV_test_mode: bool = False,
                 run_async: bool = False,
                 ext_power_supply: bool = False,
                 module_invert: bool | List[bool] = False,
                 voltage_table_selection: Optional[str] = None,
                 sequence_time_selection: Optional[str] = None) -> None:
        """
        Initialize the LIFUInterface with given parameters and store them in the class.

        Args:
            vid (int): Vendor ID of the USB device.
            tx_pid (int): Product ID for TX device.
            con_pid (int): Product ID for console device.
            baudrate (int): Communication baud rate.
            timeout (int): Read timeout in seconds.
            test_mode (bool): Enable test mode.
            run_async (bool): Enable asynchronous operation.
        """
        # Store parameters in instance variables
        self.vid = vid
        self.tx_pid = tx_pid
        self.con_pid = con_pid
        self.baudrate = baudrate
        self.timeout = timeout
        self._test_mode = TX_test_mode
        self._async_mode = run_async
        self._tx_uart = None
        self._hv_uart = None
        self.status = LIFUInterfaceStatus.STATUS_SYS_OFF

        self.voltage_table = None
        self.sequence_time = None
        self.voltage_table_selection = voltage_table_selection
        self.sequence_time_selection = sequence_time_selection

        # Create a TXDevice instance as part of the interface
        logger.debug("Initializing TX Module of LIFUInterface with VID: %s, PID: %s, baudrate: %s, timeout: %s", vid, tx_pid, baudrate, timeout)
        self._tx_uart = LIFUUart(vid=vid, pid=tx_pid, baudrate=baudrate, timeout=timeout, desc="TX", demo_mode=TX_test_mode, async_mode=run_async)
        self.txdevice = TxDevice(uart=self._tx_uart, module_invert=module_invert)

        if ext_power_supply:
            logger.debug("External power supply selected, skipping HVController initialization.")
            self.hvcontroller = None
        else:
            # Create a LIFUHVController instance as part of the interface
            logger.debug("Initializing Console of LIFUInterface with VID: %s, PID: %s, baudrate: %s, timeout: %s", vid, con_pid, baudrate, timeout)
            self._hv_uart = LIFUUart(vid=vid, pid=con_pid, baudrate=baudrate, timeout=timeout, desc="HV", demo_mode=HV_test_mode, async_mode=run_async)
            self.hvcontroller = HVController(uart=self._hv_uart)

        # Connect signals to internal handlers
        if self._async_mode:
            self._tx_uart.signal_connect.connect(self.signal_connect.emit)
            self._tx_uart.signal_disconnect.connect(self.signal_disconnect.emit)
            self._tx_uart.signal_data_received.connect(self.signal_data_received.emit)
            self._hv_uart.signal_connect.connect(self.signal_connect.emit)
            self._hv_uart.signal_disconnect.connect(self.signal_disconnect.emit)
            self._hv_uart.signal_data_received.connect(self.signal_data_received.emit)

    # Temporary fix for hardware variations between EVT0 and EVT2
    def _resolve_voltage_chart_evt_version(self, voltage_table: str) -> list[list[int]]:
        if voltage_table is None:
            try:
                evt_version = "evt0" if self.hvcontroller.get_version().startswith("v1.1") else "evt2"
            except Exception as e:
                logger.error("Error getting console version: %s", e)
                raise e
        else:
            evt_version = voltage_table.lower()
            if evt_version not in MAX_VOLTAGE_BY_DUTY_CYCLE_AND_SEQUENCE_TIME:
                raise ValueError(f"Invalid voltage_table option '{voltage_table}'. Valid options are: {tuple(MAX_VOLTAGE_BY_DUTY_CYCLE_AND_SEQUENCE_TIME.keys())}")

        return MAX_VOLTAGE_BY_DUTY_CYCLE_AND_SEQUENCE_TIME[evt_version]

    # Restrict sequence time options for users vs QA
    def _resolve_max_sequence_time_set(self, sequence_time: str) -> list[int]:
        if sequence_time is None:
            return REF_MAX_SEQUENCE_TIMES["default"]
        else:
            sequence_time = sequence_time.lower()
            if sequence_time not in REF_MAX_SEQUENCE_TIMES:
                raise ValueError(f"Invalid sequence_time option '{sequence_time}'. Valid options are: {tuple(REF_MAX_SEQUENCE_TIMES.keys())}")
            return REF_MAX_SEQUENCE_TIMES[sequence_time]

    async def start_monitoring(self, interval: int = 1) -> None:
        """Start monitoring for USB device connections."""
        try:
            if self._hv_uart is not None:
                await asyncio.gather(
                    self._tx_uart.monitor_usb_status(interval),
                    self._hv_uart.monitor_usb_status(interval)
                )
            else:
                await self._tx_uart.monitor_usb_status(interval)

        except Exception as e:
            logger.error("Error starting monitoring: %s", e)
            raise e

    def stop_monitoring(self) -> None:
        """Stop monitoring for USB device connections."""
        try:
            if self._tx_uart is not None:
                self._tx_uart.stop_monitoring()
            if self._hv_uart is not None:
                self._hv_uart.stop_monitoring()
        except Exception as e:
            logger.error("Error stopping monitoring: %s", e)
            raise e

    def is_device_connected(self) -> tuple:
        """
        Check if the device is currently connected.

        Returns:
            tuple: (tx_connected, hv_connected)
        """
        tx_connected = self.txdevice.is_connected()
        if self.hvcontroller is None:
            hv_connected = False
        else:
            hv_connected = self.hvcontroller.is_connected()
        return tx_connected, hv_connected

    def get_max_voltage(self, solution: Solution | Dict) -> float:
        """
        Get the maximum voltage for a given solution.

        Args:
            solution (Solution | Dict): The solution to check.

        Returns:
            float: The maximum voltage for the solution.
        """
        if isinstance(solution, Solution):
            solution = solution.to_dict()

        sequence_duty_cycle = self.get_sequence_duty_cycle(solution)
        sequence_duration = self.get_sequence_duration(solution)

        # Find the index of the duty cycle in the reference list
        duty_cycles_limits = np.array(REF_MAX_DUTY_CYCLES)
        duty_cycle_index = np.where(duty_cycles_limits >= sequence_duty_cycle)[0][0]

        # Find the index of the duration in the reference list
        duration_limits = np.array(self.sequence_time)
        duration_index = np.where(duration_limits >= sequence_duration)[0][0]

        # Return the maximum voltage for the given duty cycle and duration
        return self.voltage_table[duty_cycle_index][duration_index]

    def get_max_voltage_table(self) -> pd.DataFrame:
        """
        Get a table of the maximum voltages for different duty cycles and sequence times.

        Returns:
            pd.DataFrame: A DataFrame containing the maximum voltages.
        """
        data = {
            "Duty Cycle (%)": [f"<={100 * dc:0.1f}%" for dc in REF_MAX_DUTY_CYCLES],
            }
        for i, duration in enumerate(self.sequence_time):
            col_name = f"<={duration // 60} min"
            data[col_name] = [
                self.voltage_table[j][i] for j in range(len(REF_MAX_DUTY_CYCLES))
            ]
        max_voltage =  pd.DataFrame(data).set_index("Duty Cycle (%)")
        max_voltage.Name = "Maximum Voltage (V)"
        max_voltage.Description = "This table shows the maximum voltage for different duty cycles and sequence times."
        return max_voltage

    def check_solution(self, solution: Solution | Dict) -> None:
        """
        Check if the solution is valid.
        Args:
            solution (Solution | Dict): The solution to check.
        Raises:
            ValueError: If the solution is invalid.
        """
        if isinstance(solution, Solution):
            solution = solution.to_dict()

        self.voltage_table = self._resolve_voltage_chart_evt_version(self.voltage_table_selection)
        self.sequence_time = self._resolve_max_sequence_time_set(self.sequence_time_selection)
        sequence_duty_cycle = self.get_sequence_duty_cycle(solution)
        duty_cycles_limits = np.array(REF_MAX_DUTY_CYCLES)
        if sequence_duty_cycle > duty_cycles_limits.max():
            raise ValueError(f"Sequence duty cycle ({100*sequence_duty_cycle:0.1f} %) exceeds maximum allowed duty cycle ({100*duty_cycles_limits.max():0.1f} %).")
        duty_cycle_index = np.where(duty_cycles_limits >= sequence_duty_cycle)[0][0]

        sequence_duration = self.get_sequence_duration(solution)
        duration_limits = np.array(self.sequence_time)
        if sequence_duration > duration_limits.max():
            raise ValueError(f"Sequence duration ({sequence_duration:0.0f} s) exceeds maximum allowed duration ({duration_limits.max()} s).")
        duration_index = np.where(duration_limits >= sequence_duration)[0][0]

        max_voltage = self.voltage_table[duty_cycle_index][duration_index]
        if solution['voltage'] > max_voltage:
            raise ValueError(f"Voltage ({solution['voltage']:0.1f}V) exceeds maximum allowed voltage ({max_voltage:0.1f}V) for duty cycle ({100*sequence_duty_cycle:0.1f} <= {100*duty_cycles_limits[duty_cycle_index]}%) and sequence time ({sequence_duration:0.0f} <= {duration_limits[duration_index]}s).")

    def get_sequence_duty_cycle(self, solution: Solution | Dict) -> float:
        """
        Get the duty cycle of the sequence in the solution.

        Args:
            solution (Solution | Dict): The solution to check.

        Returns:
            float: The duty cycle of the sequence.
        """
        if isinstance(solution, Solution):
            solution = solution.to_dict()

        if solution['sequence']['pulse_train_interval'] == 0:
            return solution['pulse']['duration'] / solution['sequence']['pulse_interval']
        else:
            return (solution['pulse']['duration'] * solution['sequence']['pulse_count']) / solution['sequence']['pulse_train_interval']

    def get_sequence_duration(self, solution: Solution | Dict) -> float:
        """
        Get the duration of the sequence in the solution.

        Args:
            solution (Solution | Dict): The solution to check.

        Returns:
            float: The duration of the sequence.
        """
        if isinstance(solution, Solution):
            solution = solution.to_dict()

        if solution['sequence']['pulse_train_interval'] == 0:
            return solution['sequence']['pulse_interval'] * solution['sequence']['pulse_count'] * solution['sequence']['pulse_train_count']
        else:
            return solution['sequence']['pulse_train_interval'] * solution['sequence']['pulse_train_count']

    def set_module_invert(self, module_invert: bool | List[bool]) -> None:
        if self.txdevice is not None:
            self.txdevice.set_module_invert(module_invert)

    def set_solution(self,
                     solution: Solution | Dict,
                     profile_index:int=1,
                     profile_increment:bool=True,
                     trigger_mode: TriggerModeOpts = "sequence",
                     ) -> None:
        """
        Load a solution to the device.

        Args:
            solution (Solution): The solution to load.
            profile_index (int): The profile index to load the solution to (defaults to 0)
            profile_increment (bool): Increment the profile index
            trigger_mode (TriggerModeOpts): The trigger mode to use (defaults to "sequence")
            module_invert (List[bool]|bool): Invert the signal on all modules (singleton) or specific modules (list) (defaults to False)
        """
        if isinstance(solution, Solution):
            solution = solution.to_dict()

        self.check_solution(solution)

        if "transducer" in solution and solution["transducer"] is not None and "module_invert" in solution["transducer"]:
            self.txdevice.set_module_invert(solution["transducer"]["module_invert"])
        else:
            self.txdevice.set_module_invert(False)

        self.set_status(LIFUInterfaceStatus.STATUS_PROGRAMMING)

        if "name" in solution:
            solution_name = solution["name"]
            solution_name = f'Solution "{solution_name}"'
        else:
            solution_name = "Solution"

        voltage = solution['voltage']
        logger.info("Loading %s...", solution_name)
        # Convert solution data and send to the device
        self.txdevice.set_solution(
                pulse = solution['pulse'],
                delays = solution['delays'],
                apodizations= solution['apodizations'],
                sequence= solution['sequence'],
                profile_index=profile_index,
                profile_increment=profile_increment,
                trigger_mode=trigger_mode
        )
        self.set_status(LIFUInterfaceStatus.STATUS_READY)

        if self.hvcontroller is not None:
            logger.info(f"Setting HV to {voltage} V...")
            self.hvcontroller.set_voltage(voltage)

        logger.info("%s loaded successfully.", solution_name)

    def start_sonication(self) -> bool:
        """
        Start sonication.

        Sets the device to a running state and sends a start command if necessary.
        """
        try:
            if self._test_mode:
                return True

            if self.hvcontroller is not None:
                logger.info("Turn ON HV")
                bHvOn = self.hvcontroller.turn_hv_on()
            else:
                logger.info("Using external power supply, HV will not be turned ON.")
                bHvOn = True

            if self._async_mode:
                self.txdevice.async_mode(True)

            logger.info("Start Sonication")
            # Send the solution data to the device
            bTriggerOn = self.txdevice.start_trigger()

            if bTriggerOn and bHvOn:
                logger.info("Sonication started successfully.")
                self.set_status(LIFUInterfaceStatus.STATUS_RUNNING)
                return True
            else:
                logger.error("Failed to start sonication.")
                return False

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Error Starting sonication: %s", e)
            raise e

    def set_status(self, status: LIFUInterfaceStatus) -> None:
        """
        Set the device status.

        Args:
            status (LIFUInterfaceStatus): The status to set.
        """
        logger.info("Setting device status to %s", status.name)
        self.status = status

    def get_status(self) -> LIFUInterfaceStatus:
        """
        Query the device status.

        Returns:
            int: The device status.
        """
        if self._test_mode:
            return LIFUInterfaceStatus.STATUS_READY

        return self.status

    def stop_sonication(self) -> bool:
        """
        Stop sonication.

        Stops the current sonication process.
        """
        try:
            if self._test_mode:
                return True

            logger.info("Stop Sonication")
            # Send the solution data to the device
            bTriggerOff = self.txdevice.stop_trigger()

            if self.hvcontroller is not None:
                logger.info("Turn OFF HV")
                bHvOff = self.hvcontroller.turn_hv_off()
            else:
                logger.info("Using external power supply, HV will not be turned OFF.")
                bHvOff = True

            if self._async_mode:
                self.txdevice.async_mode(False)

            if bTriggerOff and bHvOff:
                logger.info("Sonication stopped successfully.")
                return True
            else:
                logger.error("Failed to stop sonication.")
                return False

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Error Stopping sonication: %s", e)
            raise e

    def close(self):
        self.stop_monitoring()
        if self.txdevice:
            self.txdevice.close()
        if self.hvcontroller:
            self.hvcontroller.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @staticmethod
    def get_sdk_version() -> str:
        return openlifu.__version__



from __future__ import annotations

import sys
import time

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_async_mode.py

def main():
    print("Starting LIFU Async Test Script...")
    interface = LIFUInterface()
    tx_connected, hv_connected = interface.is_device_connected()

    if not tx_connected and not hv_connected:
        print("✅ LIFU Console not connected.")
        sys.exit(1)

    if not tx_connected:
        print("TX device not connected. Attempting to turn on 12V...")
        interface.hvcontroller.turn_12v_on()
        time.sleep(2)

        interface.stop_monitoring()
        del interface
        time.sleep(3)

        print("Reinitializing LIFU interface after powering 12V...")
        interface = LIFUInterface()
        tx_connected, hv_connected = interface.is_device_connected()

        if tx_connected and hv_connected:
            print("✅ LIFU Device fully connected.")
        else:
            print("❌ LIFU Device NOT fully connected.")
            print(f"  TX Connected: {tx_connected}")
            print(f"  HV Connected: {hv_connected}")
            sys.exit(1)

    print("Ping the device")
    if not interface.txdevice.ping():
        print("❌ Failed comms with txdevice.")
        sys.exit(1)

    version = interface.txdevice.get_version()
    print(f"Version: {version}")

    curr_mode = interface.txdevice.async_mode()
    print(f"Current Async Mode: {curr_mode}")
    time.sleep(1)
    if curr_mode:
        print("Async mode is already enabled.")
    else:
        print("Enabling Async Mode...")
        interface.txdevice.async_mode(True)
        time.sleep(1)
        curr_mode = interface.txdevice.async_mode()
        print(f"Async Mode Enabled: {curr_mode}")
    time.sleep(1)
    print("Disabling Async Mode...")
    interface.txdevice.async_mode(False)
    time.sleep(1)
    curr_mode = interface.txdevice.async_mode()
    print(f"Async Mode Enabled: {curr_mode}")

if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import os
import sys
import time
from pathlib import Path

if os.name == 'nt':
    pass
else:
    pass

import numpy as np

from openlifu.bf.pulse import Pulse
from openlifu.bf.sequence import Sequence
from openlifu.db import Database
from openlifu.geo import Point
from openlifu.io.LIFUInterface import LIFUInterface
from openlifu.plan.solution import Solution

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_watertank.py

"""
Test script to automate:
1. Connect to the device.
2. Test HVController: Turn HV on/off and check voltage.
3. Test Device functionality.
"""

# Configure logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Prevent duplicate handlers and cluttered terminal output
if not logger.hasHandlers():
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))
    logger.addHandler(handler)
    logger.propagate = False

log_interval = 1  # seconds; you can adjust this variable as needed

# set focus
xInput = 0
yInput = 0
zInput = 50

frequency_kHz = 400 # Frequency in kHz
voltage = 20.0 # Voltage in Volts
duration_msec = 5/400 # Pulse Duration in milliseconds
interval_msec = 20 # Pulse Repetition Interval in milliseconds
num_modules = 2 # Number of modules in the system

use_external_power_supply = False # Select whether to use console or power supply

console_shutoff_temp_C = 70.0 # Console shutoff temperature in Celsius
tx_shutoff_temp_C = 70.0 # TX device shutoff temperature in Celsius
ambient_shutoff_temp_C = 70.0 # Ambient shutoff temperature in Celsius

#TODO: script_timeout_minutes = 30 # Prevent unintentionally leaving unit on for too long
#TODO: log_temp_to_csv_file = True # Log readings to only terminal or both terminal and CSV file

# Fail-safe parameters if the temperature jumps too fast
rapid_temp_shutoff_C = 40 # Cutoff temperature in Celsius if it jumps too fast
rapid_temp_shutoff_seconds = 5 # Time in seconds to reach rapid temperature shutoff
rapid_temp_increase_per_second_shutoff_C = 3 # Rapid temperature climbing shutoff in Celsius

peak_to_peak_voltage = voltage * 2 # Peak to peak voltage for the pulse

here = Path(__file__).parent.resolve()
db_path = here / ".." / "db_dvc"
db = Database(db_path)
arr = db.load_transducer(f"openlifu_{num_modules}x400_evt1_005")
arr.sort_by_pin()

target = Point(position=(xInput,yInput,zInput), units="mm")
focus = target.get_position(units="mm")
distances = np.sqrt(np.sum((focus - arr.get_positions(units="mm"))**2, 1)).reshape(1,-1)
tof = distances*1e-3 / 1500
delays = tof.max() - tof
print(f"TOF Max = {tof.max()*1e6} us")

apodizations = np.ones((1, arr.numelements()))
#active_element = 25
#active_element = np.arange(65,128)
#apodizations = np.zeros((1, arr.numelements()))
#apodizations[:, active_element-1] = 1

logger.info("Starting LIFU Test Script...")
interface = LIFUInterface(ext_power_supply=use_external_power_supply)
tx_connected, hv_connected = interface.is_device_connected()

if not use_external_power_supply and not tx_connected:
    logger.warning("TX device not connected. Attempting to turn on 12V...")
    interface.hvcontroller.turn_12v_on()

    # Give time for the TX device to power up and enumerate over USB
    time.sleep(2)

    # Cleanup and recreate interface to reinitialize USB devices
    interface.stop_monitoring()
    del interface
    time.sleep(1)  # Short delay before recreating

    logger.info("Reinitializing LIFU interface after powering 12V...")
    interface =  LIFUInterface(ext_power_supply=use_external_power_supply)

    # Re-check connection
    tx_connected, hv_connected = interface.is_device_connected()

if not use_external_power_supply:
    if hv_connected:
        logger.info(f"  HV Connected: {hv_connected}")
    else:
        logger.error("❌ HV NOT fully connected.")
        sys.exit(1)
else:
    logger.info("  Using external power supply")

if tx_connected:
    logger.info(f"  TX Connected: {tx_connected}")
    logger.info("✅ LIFU Device fully connected.")
else:
    logger.error("❌ TX NOT fully connected.")
    sys.exit(1)

stop_logging = False  # flag to signal the logging thread to stop

# Verify communication with the devices
if not interface.txdevice.ping():
    logger.error("Failed to ping the transmitter device.")
    sys.exit(1)

if not use_external_power_supply and not interface.hvcontroller.ping():
    logger.error("Failed to ping the console device.")
    sys.exit(1)

if not use_external_power_supply:
    try:
        console_firmware_version = interface.hvcontroller.get_version()
        logger.info(f"Console Firmware Version: {console_firmware_version}")
    except Exception as e:
        logger.error(f"Error querying console firmware version: {e}")
try:
    tx_firmware_version = interface.txdevice.get_version()
    logger.info(f"TX Firmware Version: {tx_firmware_version}")
except Exception as e:
    logger.error(f"Error querying TX firmware version: {e}")

logger.info("Enumerate TX7332 chips")
num_tx_devices = interface.txdevice.enum_tx7332_devices()
if num_tx_devices == 0:
    raise ValueError("No TX7332 devices found.")
elif num_tx_devices == num_modules*2:
    logger.info(f"Number of TX7332 devices found: {num_tx_devices}")
    numelements = 32*num_tx_devices
else:
    raise Exception(f"Number of TX7332 devices found: {num_tx_devices} != 2x{num_modules}")

logger.info(f'Apodizations: {apodizations}')
logger.info(f'Delays: {delays}')

pulse = Pulse(frequency=frequency_kHz*1e3, duration=duration_msec*1e-3)

sequence = Sequence(
    pulse_interval=interval_msec*1e-3,
    pulse_count=int(60/(interval_msec*1e-3)),
    pulse_train_interval=0,
    pulse_train_count=1
)

pin_order = np.argsort([el.pin for el in arr.elements])
solution = Solution(
    delays = delays[:, pin_order],
    apodizations = apodizations[:, pin_order],
    transducer=arr,
    pulse = pulse,
    voltage=voltage,
    sequence = sequence
)
profile_index = 1
profile_increment = True
trigger_mode = "single"

if use_external_power_supply:
    logger.info(f"Using external power supply. Ensure HV is turned on and set to {voltage}V before starting the trigger.")

interface.set_solution(
    solution=solution,
    profile_index=profile_index,
    profile_increment=profile_increment,
    trigger_mode=trigger_mode)

logger.info("Get Trigger")
trigger_setting = interface.txdevice.get_trigger_json()
if trigger_setting:
    logger.info(f"Trigger Setting: {trigger_setting}")
else:
    logger.error("Failed to get trigger setting.")
    sys.exit(1)

duty_cycle = int((duration_msec/interval_msec) * 100)
if duty_cycle > 50:
    logger.warning("❗❗ Duty cycle is above 50% ❗❗")

logger.info(f"User parameters set: \n\
    Module Invert: {arr.module_invert}\n\
    Frequency: {frequency_kHz}kHz\n\
    Voltage Per Rail: {voltage}V\n\
    Voltage Peak to Peak: {peak_to_peak_voltage}V\n\
    Duration: {duration_msec}ms\n\
    Interval: {interval_msec}ms\n\
    Duty Cycle: {duty_cycle}%\n\
    Use External Power Supply: {use_external_power_supply}\n\
    Initial Temp Safety Shutoff: Increase to {rapid_temp_shutoff_C}°C within {rapid_temp_shutoff_seconds}s of starting.\n\
    General Temp Safety Shutoff: Increase of {rapid_temp_increase_per_second_shutoff_C}°C within {log_interval}s at any point.\n")

def turn_off_console_and_tx():
    if not use_external_power_supply:
        logger.info("Attempting to turn off High Voltage...")
        interface.hvcontroller.turn_hv_off()
        if interface.hvcontroller.get_hv_status():
            logger.error("High Voltage is still on.")
        else:
            logger.info("High Voltage successfully turned off.")

        logger.info("Attempting to turn off 12V...")
        interface.hvcontroller.turn_12v_off()
        if interface.hvcontroller.get_12v_status():
            logger.error("12V is still on.")
        else:
            logger.info("12V successfully turned off.")

logger.info("turning HV ON")
interface.hvcontroller.turn_hv_on()
s = input("Press any key to start")
logger.info("Sending Single Trigger...")
interface.txdevice.start_trigger()
time.sleep(0.1)
interface.txdevice.stop_trigger()
interface.hvcontroller.turn_hv_off()
turn_off_console_and_tx()
logger.info("Finished")



from __future__ import annotations

import base64
import json
import tempfile
from dataclasses import asdict, dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Annotated, Dict, List, Tuple

import numpy as np
import xarray as xa

from openlifu.bf import Pulse, Sequence
from openlifu.bf.focal_patterns import FocalPattern
from openlifu.geo import Point
from openlifu.plan.param_constraint import ParameterConstraint
from openlifu.plan.solution_analysis import (
    SolutionAnalysis,
    SolutionAnalysisOptions,
    find_centroid,
    get_beamwidth,
    get_mask,
)
from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.util.json import PYFUSEncoder
from openlifu.util.units import getunitconversion, rescale_coords, rescale_data_arr
from openlifu.xdc import Transducer


def _construct_nc_filepath_from_json_filepath(json_filepath:Path) -> Path:
    """Construct a default filepath to netCDF file given filepath to associated solution json file."""
    nc_filename = json_filepath.name.split(".")[0] + ".nc"
    nc_filepath = json_filepath.parent / nc_filename
    return nc_filepath


@dataclass
class Solution:
    """
    A sonication solution resulting from beamforming and running a simulation.
    """

    id: Annotated[str, OpenLIFUFieldData("Solution ID", "ID of this solution")] = "solution"  # the *solution* id, a concept that did not exist in the matlab software
    """ID of this solution"""

    name: Annotated[str, OpenLIFUFieldData("Solution name", "Name of this solution")] = "Solution"
    """Name of this solution"""

    protocol_id: Annotated[str | None, OpenLIFUFieldData("Protocol ID", "ID of the protocol that was used when generating this solution")] = None  # this used to be called plan_id in the matlab code
    """ID of the protocol that was used when generating this solution"""

    transducer: Annotated[Transducer | None, OpenLIFUFieldData("Transducer", "Transducer used when generating this solution")] = None
    """Transducer used when generating this solution"""

    date_created: Annotated[datetime, OpenLIFUFieldData("Creation date", "Solution creation time")] = field(default_factory=datetime.now)
    """Solution creation time"""

    description: Annotated[str, OpenLIFUFieldData("Description", "Description of this solution")] = ""
    """Description of this solution"""

    delays: Annotated[np.ndarray | None, OpenLIFUFieldData("Delays", "Vectors of time delays to steer the beam. Shape is (number of foci, number of transducer elements).")] = None
    """Vectors of time delays to steer the beam. Shape is (number of foci, number of transducer elements)."""

    apodizations: Annotated[np.ndarray | None, OpenLIFUFieldData("Apodizations", "Vectors of apodizations to steer the beam. Shape is (number of foci, number of transducer elements).")] = None
    """Vectors of apodizations to steer the beam. Shape is (number of foci, number of transducer elements)."""

    pulse: Annotated[Pulse, OpenLIFUFieldData("Pulse", "Pulse to send to the transducer when running sonication")] = field(default_factory=Pulse)
    """Pulse to send to the transducer when running sonication"""

    voltage: Annotated[float, OpenLIFUFieldData("Voltage", "Voltage to use when running sonication. This is the voltage that will be set on the HV power supply.")] = 1.0
    """Voltage to use when running sonication. This is the voltage that will be set on the HV power supply."""

    sequence: Annotated[Sequence, OpenLIFUFieldData("Pulse sequence", "Pulse sequence to use when running sonication")] = field(default_factory=Sequence)
    """Pulse sequence to use when running sonication"""

    foci: Annotated[List[Point], OpenLIFUFieldData("Foci", "Points that are focused on in this Solution due to the focal pattern around the target. Each item in this list is a unique point from the focal pattern, and the pulse sequence is what determines how many times each point will be used.")] = field(default_factory=list)
    """Points that are focused on in this Solution due to the focal pattern around the target.
    Each item in this list is a unique point from the focal pattern, and the pulse sequence is
    what determines how many times each point will be used.
    """

    # there was "target_id" in the matlab software, but here we do not have the concept of a target ID.
    # I believe this was only needed in the matlab software because solutions were organized by target rather
    # than having their own unique solution ID. We do have unique solution IDs so it's possible we don't need
    # this target attribute at all here. Keeping it here for now just in case.
    target: Annotated[Point | None, OpenLIFUFieldData("Target point", "The ultimate target of this sonication. This sonication solution is focused on one focal point in a pattern that is centered on this target.")] = None
    """The ultimate target of this sonication. This sonication solution is focused on one focal point
    in a pattern that is centered on this target."""

    # In the matlab code the simulation result was saved as a separate .mat file.
    # Here we include it as an xarray dataset.
    simulation_result: Annotated[xa.Dataset, OpenLIFUFieldData("Simulation result", "The xarray Dataset of simulation results")] = field(default_factory=xa.Dataset)
    """The xarray Dataset of simulation results"""

    approved: Annotated[bool, OpenLIFUFieldData("Approved?", "Approval state of this solution as a sonication plan. `True` means the user has provided some kind of confirmation that the solution is safe and acceptable to be executed.")] = False
    """Approval state of this solution as a sonication plan. `True` means the user has provided some
    kind of confirmation that the solution is safe and acceptable to be executed."""

    def __post_init__(self):
        if self.delays is not None:
            self.delays = np.array(self.delays, ndmin=2)
        if self.apodizations is not None:
            self.apodizations = np.array(self.apodizations, ndmin=2)

        if self.pulse.frequency <= 0:
            raise ValueError("Pulse frequency must be positive")
        if self.voltage <= 0:
            raise ValueError("Voltage must be positive")
        if self.sequence.pulse_interval <= 0:
            raise ValueError("Pulse interval must be positive")
        if self.sequence.pulse_count <= 0:
            raise ValueError("Pulse count must be positive")
        if self.sequence.pulse_train_interval < 0:
            raise ValueError("Pulse train interval must be non-negative")
        elif (self.sequence.pulse_train_interval > 0) and (self.sequence.pulse_train_interval < (self.sequence.pulse_interval * self.sequence.pulse_count)):
            raise ValueError("Pulse train interval must be greater than or equal to the total pulse interval")
        if self.sequence.pulse_train_count <= 0:
            raise ValueError("Pulse train count must be positive")
        if len(self.foci)>0 and self.delays is not None and self.delays.shape[0] != len(self.foci):
            raise ValueError(f"Delays number of foci ({self.delays.shape[0]}) does not match number of foci ({len(self.foci)})")
        if len(self.foci)>0 and self.apodizations is not None and self.apodizations.shape[0] != len(self.foci):
            raise ValueError(f"Apodizations number of foci ({self.apodizations.shape[0]}) does not match number of foci ({len(self.foci)})")
        if self.delays is not None and self.apodizations is not None:
            if self.apodizations.shape[0] != self.delays.shape[0]:
                raise ValueError(f"Apodizations number of foci ({self.apodizations.shape[0]}) does not match delays number of foci ({self.delays.shape[0]})")
            if self.apodizations.shape[1] != self.delays.shape[1]:
                raise ValueError(f"Apodizations number of elements {self.apodizations.shape[1]} does not match delays shape ({self.delays.shape[1]})")


    def num_foci(self) -> int:
        """Get the number of foci"""
        return len(self.foci)

    def analyze(self,
                options: SolutionAnalysisOptions = SolutionAnalysisOptions(),
                param_constraints: Dict[str,ParameterConstraint] | None = None) -> SolutionAnalysis:
        """Analyzes the treatment solution.

        Args:
            options: A struct for solution analysis options.
            param_constraints: A dictionary of parameter constraints to apply to the analysis.
                The keys are the parameter names and the values are the ParameterConstraint objects.

        Returns: A struct containing the results of the analysis.
        """
        if param_constraints is None:
            param_constraints = {}
        solution_analysis = SolutionAnalysis()

        dt = 1 / (self.pulse.frequency * 20)
        t = self.pulse.calc_time(dt)
        input_signal_V = self.pulse.calc_pulse(t) * self.voltage

        pnp_MPa_all = rescale_data_arr(rescale_coords(self.simulation_result['p_min'], options.distance_units),"MPa")
        ipa_Wcm2_all = rescale_data_arr(rescale_coords(self.simulation_result['intensity'], options.distance_units), "W/cm^2")

        if options.sidelobe_radius is np.nan:
            options.sidelobe_radius = options.mainlobe_radius


        standoff_Z = options.standoff_density * 1500
        c_tic = 40e-3  # W cm-1
        A_cm = self.transducer.get_area("cm")
        d_eq_cm = np.sqrt(4*A_cm / np.pi)
        ele_sizes_cm2 = np.array([elem.get_area("cm") for elem in self.transducer.elements])

        # xyz = np.stack(np.meshgrid(*coords, indexing="xy"), axis=-1)  #TODO: if fus.Axis is defined, coords.ndgrid(dim="z")
        # z_mask = xyz[..., -1] >= options.sidelobe_zmin  #TODO: probably wrong here, should be z{1}>=options.sidelobe_zmin;

        solution_analysis.duty_cycle_pulse_train_pct = self.get_pulsetrain_dutycycle()*100
        solution_analysis.duty_cycle_sequence_pct = self.get_sequence_dutycycle()*100
        if self.sequence.pulse_train_interval == 0:
            solution_analysis.sequence_duration_s = float(self.sequence.pulse_interval * self.sequence.pulse_count * self.sequence.pulse_train_count)
        else:
            solution_analysis.sequence_duration_s = float(self.sequence.pulse_train_interval * self.sequence.pulse_train_count)
        ita_mWcm2 = rescale_coords(self.get_ita(units="mW/cm^2"), options.distance_units)

        power_W = np.zeros(self.num_foci())
        TIC = np.zeros(self.num_foci())
        for focus_index in range(self.num_foci()):
            pnp_MPa = pnp_MPa_all.isel(focal_point_index=focus_index)
            ipa_Wcm2 = ipa_Wcm2_all.isel(focal_point_index=focus_index)
            focus = self.foci[focus_index].get_position(units=options.distance_units)
            focus_mm = self.foci[focus_index].get_position(units="mm")
            solution_analysis.target_position_lat_mm += [focus_mm[0]]
            solution_analysis.target_position_ele_mm += [focus_mm[1]]
            solution_analysis.target_position_ax_mm += [focus_mm[2]]  # TODO: this should be a list, not a single value
            apodization = self.apodizations[focus_index]
            origin = self.transducer.get_effective_origin(apodizations=apodization, units=options.distance_units)

            output_signal_Pa = self.transducer.calc_output(input_signal_V, dt, delays=self.delays[focus_index, :], apod=self.apodizations[focus_index, :])
            p0_Pa = np.max(output_signal_Pa, axis=1)

            mainlobe_mask = get_mask(
                pnp_MPa,
                focus = focus,
                origin = origin,
                distance = options.mainlobe_radius,
                operator = '<',
                aspect_ratio = options.mainlobe_aspect_ratio
            )

            sidelobe_mask = get_mask(
                pnp_MPa,
                focus = focus,
                origin = origin,
                distance = options.sidelobe_radius,
                operator = '>',
                aspect_ratio=options.mainlobe_aspect_ratio
            )
            z_dim = pnp_MPa.dims[2]
            z_mask = pnp_MPa.coords[z_dim] > options.sidelobe_zmin
            sidelobe_mask = sidelobe_mask.where(z_mask, False)

            pnp_mainlobe = pnp_MPa.where(mainlobe_mask)
            pk = float(pnp_mainlobe.max())
            mainlobe_focus = find_centroid(pnp_mainlobe, pk*10**(-3/20), "mm")
            solution_analysis.focal_centroid_lat_mm += [mainlobe_focus[0]]
            solution_analysis.focal_centroid_ele_mm += [mainlobe_focus[1]]
            solution_analysis.focal_centroid_ax_mm += [mainlobe_focus[2]]

            solution_analysis.mainlobe_pnp_MPa += [pk]

            for dim, named_dim, scale in zip(pnp_MPa.dims, ("lat","ele","ax"), options.mainlobe_aspect_ratio):
                for threshdB in [3, 6]:
                    attr_name = f'beamwidth_{named_dim}_{threshdB}dB_mm'
                    bw0 = getattr(solution_analysis, attr_name)
                    cutoff = pk*10**(-threshdB / 20)
                    bw = get_beamwidth(
                        pnp_MPa,
                        focus=focus,
                        dim=dim,
                        cutoff=cutoff,
                        origin=origin,
                        min_offset=-scale*options.beamwidth_radius,
                        max_offset=scale*options.beamwidth_radius)
                    bw = getunitconversion(options.distance_units, "mm") * bw
                    bw = [*bw0, bw]
                    setattr(solution_analysis, attr_name, bw)

            solution_analysis.mainlobe_isppa_Wcm2 += [float(ipa_Wcm2.where(mainlobe_mask).max())]
            solution_analysis.mainlobe_ispta_mWcm2 += [float(ita_mWcm2.where(mainlobe_mask).max())]

            sidelobe_pnp = float(pnp_MPa.where(sidelobe_mask).max())
            sidelobe_isppa = float(ipa_Wcm2.where(sidelobe_mask).max())

            solution_analysis.sidelobe_pnp_MPa += [sidelobe_pnp]
            solution_analysis.sidelobe_isppa_Wcm2 += [sidelobe_isppa]

            # Calculate and store ratios (sidelobe / mainlobe)
            mainlobe_pnp = solution_analysis.mainlobe_pnp_MPa[-1]
            mainlobe_isppa = solution_analysis.mainlobe_isppa_Wcm2[-1]

            if mainlobe_pnp == 0:
                pressure_ratio = np.inf if sidelobe_pnp != 0 else np.nan
            else:
                pressure_ratio = sidelobe_pnp / mainlobe_pnp
            solution_analysis.sidelobe_to_mainlobe_pressure_ratio += [pressure_ratio]

            if mainlobe_isppa == 0:
                intensity_ratio = np.inf if sidelobe_isppa != 0 else np.nan
            else:
                intensity_ratio = sidelobe_isppa / mainlobe_isppa
            solution_analysis.sidelobe_to_mainlobe_intensity_ratio += [intensity_ratio]

            solution_analysis.global_pnp_MPa += [float(pnp_MPa.where(z_mask).max())]
            solution_analysis.global_isppa_Wcm2 += [float(ipa_Wcm2.where(z_mask).max())]
            i0_Wcm2 = (p0_Pa**2 / (2*standoff_Z)) * 1e-4
            i0ta_Wcm2 = i0_Wcm2 * solution_analysis.duty_cycle_sequence_pct/100
            power_W[focus_index] = np.mean(np.sum(i0ta_Wcm2 * ele_sizes_cm2 * self.apodizations[focus_index, :]))
            TIC[focus_index] = power_W[focus_index] / (d_eq_cm * c_tic)
            solution_analysis.p0_MPa += [1e-6*np.max(p0_Pa)]
        solution_analysis.global_ispta_mWcm2 = float((ita_mWcm2*z_mask).max())
        solution_analysis.MI = (np.max(solution_analysis.mainlobe_pnp_MPa)/np.sqrt(self.pulse.frequency*1e-6))
        solution_analysis.TIC = np.mean(TIC)
        solution_analysis.voltage_V = self.voltage
        solution_analysis.power_W = np.mean(power_W)

        solution_analysis.param_constraints = param_constraints
        return solution_analysis

    def compute_scaling_factors(
            self,
            focal_pattern: FocalPattern,
            analysis: SolutionAnalysis
        ) -> Tuple[np.ndarray, float, float]:
        """

        Compute the scaling factors used to re-scale the apodizations, simulation results and pulse amplitude.

        Args:
            focal_pattern: FocalPattern
            analysis: SolutionAnalysis

        Returns:
            apod_factors: A np.ndarray apodization factors
            v0: A float representing the original pulse amplitude
            v1: A float representing the new pulse amplitude
        """
        scaling_factors = np.zeros(self.num_foci())

        for i in range(self.num_foci()):
            focal_pattern_pressure_in_MPa = focal_pattern.target_pressure * getunitconversion(focal_pattern.units, "MPa")
            scaling_factors[i] = focal_pattern_pressure_in_MPa / analysis.mainlobe_pnp_MPa[i]
        max_scaling = np.max(scaling_factors)
        v0 = self.voltage
        v1 = v0 * max_scaling
        apod_factors = scaling_factors / max_scaling

        return apod_factors, v0, v1

    def scale(
            self,
            focal_pattern: FocalPattern,
            analysis_options: SolutionAnalysisOptions = SolutionAnalysisOptions()
    ) -> None:
        """
        Scale the solution in-place to match the target pressure.

        Args:
            focal_pattern: FocalPattern
            analysis_options: plan.solution.SolutionAnalysisOptions

        Returns:
            analysis_scaled: the resulting plan.solution.SolutionAnalysis from scaled solution
        """
        analysis = self.analyze(options=analysis_options)

        apod_factors, v0, v1 = self.compute_scaling_factors(focal_pattern, analysis)

        for i in range(self.num_foci()):
            scaling = v1/v0*apod_factors[i]
            self.simulation_result['p_min'][i].data *= scaling
            self.simulation_result['p_max'][i].data *= scaling
            self.simulation_result['intensity'][i].data *= scaling**2
            self.apodizations[i] = self.apodizations[i]*apod_factors[i]
        self.voltage = v1

    def get_pulsetrain_dutycycle(self) -> float:
        """
        Compute the pulse train dutycycle given a sequence.

        Returns:
            A float.
        """
        pulsetrain_dutycycle = min(1., self.pulse.duration / self.sequence.pulse_interval)

        return pulsetrain_dutycycle

    def get_sequence_dutycycle(self) -> float:
        """
        Compute the overall sequence dutycycle given a sequence.

        Returns:
            A float.
        """
        if self.sequence.pulse_train_interval == 0:
            between_pulsetrain_duty_cycle = 1
        else:
            between_pulsetrain_duty_cycle = (self.sequence.pulse_count * self.sequence.pulse_interval) / self.sequence.pulse_train_interval
        sequence_duty_cycle = self.get_pulsetrain_dutycycle() * between_pulsetrain_duty_cycle
        return sequence_duty_cycle

    def get_ita(self, units: str = "mW/cm^2") -> xa.DataArray:
        """
        Calculate the intensity-time-area product for a treatment solution.

        Args:
            output: A struct for simulation results from the treatment.
            units: str
                Target units. Default "mW/cm^2".

        Returns:
            A Solution instance with the calculated intensity value.
        """
        intensity_scaled = rescale_data_arr(self.simulation_result['intensity'], units)
        pulsetrain_dutycycle = self.get_pulsetrain_dutycycle()
        treatment_dutycycle = self.get_sequence_dutycycle()
        pulse_seq = (np.arange(self.sequence.pulse_count) - 1) % self.num_foci() + 1
        counts = np.zeros((1, 1, 1, self.num_foci()))
        for i in range(self.num_foci()):
            counts[0, 0, 0, i] = np.sum(pulse_seq == (i+1))
        intensity = intensity_scaled.copy(deep=True)
        isppa_avg = np.sum(np.expand_dims(intensity.data, axis=-1) * counts, axis=-1) / np.sum(counts)
        intensity.data = isppa_avg * pulsetrain_dutycycle * treatment_dutycycle

        return intensity

    def to_dict(self, include_simulation_data: bool = False) -> dict:
        """Serialize a Solution to a dictionary

        Args:
            include_simulation_data: if enabled then large simulation data arrays are included in the dict,
                otherwise they are excluded.

        Returns: A dictionary representing the complete Solution object.
        """
        solution_dict = asdict(self)

        if not include_simulation_data:
            solution_dict.pop('simulation_result')

        return solution_dict

    def to_json(self, include_simulation_data: bool, compact: bool) -> str:
        """Serialize a Solution to a json string

        Args:
            include_array_data: if enabled then large simulation data arrays are serialized somehow into the json,
                so that they can be recovered via `from_json` alone. otherwise they are excluded.
            compact: if enabled then the string is compact (not pretty). Disable for pretty.

        Returns: A json string representing the complete Solution object.
        """
        solution_dict = asdict(self)

        if include_simulation_data:
            # Serialize xarray dataset into a string
            with tempfile.NamedTemporaryFile(suffix=".nc", delete=False) as tmp:
                tmp_path = Path(tmp.name)
            try:
                self.simulation_result.to_netcdf(tmp_path, engine='scipy')
                # (to_netcdf can write to general byte stream instead of a file, but its behavior is changing
                # in a way that will likely break how we do it in only some python versions,
                # so writing to a tempfile is a fool proof solution across versions until to_netcdf stabilizes in its behavior)
                raw_bytes = tmp_path.read_bytes()
            finally:
                tmp_path.unlink(missing_ok=True)
            solution_dict['simulation_result'] = base64.b64encode(raw_bytes).decode('utf-8')
        else:
            solution_dict.pop('simulation_result')

        if compact:
            return json.dumps(solution_dict, separators=(',', ':'), cls=PYFUSEncoder)
        else:
            return json.dumps(solution_dict, indent=4, cls=PYFUSEncoder)

    @staticmethod
    def from_dict(solution_dict: dict) -> Solution:
        """Load a Solution from a dictionary.

        Args:
            solution_dict: The dictionary containing the solution data.

        Returns: The new Solution object.
        """
        # Convert the dictionary back into a Solution object
        solution_dict["date_created"] = datetime.fromisoformat(solution_dict["date_created"])
        if solution_dict["delays"] is not None:
            solution_dict["delays"] = np.array(solution_dict["delays"])
        if solution_dict["apodizations"] is not None:
            solution_dict["apodizations"] = np.array(solution_dict["apodizations"], ndmin=2)
            solution_dict["apodizations"] = np.array(solution_dict["apodizations"], ndmin=2)
        if solution_dict["transducer"] is not None:
            solution_dict["transducer"] = Transducer.from_dict(solution_dict["transducer"])
        solution_dict["pulse"] = Pulse.from_dict(solution_dict["pulse"])
        solution_dict["sequence"] = Sequence.from_dict(solution_dict["sequence"])
        solution_dict["foci"] = [
            Point.from_dict(focus_dict)
            for focus_dict in solution_dict["foci"]
        ]
        if solution_dict["target"] is not None:
            solution_dict["target"] = Point.from_dict(solution_dict["target"])

        if "simulation_result" in solution_dict and isinstance(solution_dict["simulation_result"], str):
            solution_dict["simulation_result"] = xa.open_dataset(
                base64.b64decode(
                    solution_dict["simulation_result"].encode('utf-8')
                ),
                engine='scipy',
            )

        return Solution(**solution_dict)


    @staticmethod
    def from_json(json_string : str, simulation_result: xa.Dataset | None=None) -> Solution:
        """Load a Solution from a json string.

        Args:
            json_string: the json string defining the Solution
            simulation_result: the simulation result arrays to use. If the json string has this then it will
                be read from the json string and it should not be provided in this argument.

        Returns: The new Solution object.
        """
        solution_dict = json.loads(json_string)
        if simulation_result is not None:
            if "simulation_result" in solution_dict:
                raise ValueError(
                    "A simulation result was provided while the json string already contains `simulation_result`. "
                    "Unclear which to use!"
                )
            solution_dict["simulation_result"] = simulation_result

        return Solution.from_dict(solution_dict)

    def to_files(self, json_filepath:Path, nc_filepath:Path | None=None) -> None:
        """Save the solution to json and netCDF files.

        json_filepath: where to save the json file with all data except the simulation results dataset
        nc_filepath: where to save the netCDF file containing the simulation results.
            If None then it will be saved in the same directory as the json file, with the same name but with
            the extension *.nc
        """
        if nc_filepath is None:
            nc_filepath = _construct_nc_filepath_from_json_filepath(json_filepath)
        json_filepath.parent.mkdir(parents=True, exist_ok=True)
        nc_filepath.parent.mkdir(parents=True, exist_ok=True)
        with json_filepath.open("w") as json_file:
            json_file.write(
                self.to_json(include_simulation_data=False, compact=False)
            )
        self.simulation_result.to_netcdf(nc_filepath, engine='h5netcdf')

    @staticmethod
    def from_files(json_filepath:Path, nc_filepath:Path | None=None):
        """Read solution from json and netCDF files.

        json_filepath: solution json file location, containing all data except the simulation results dataset
        nc_filepath: netCDF file location, containing the simulation results.
            If None then it will be assumed to be saved in the same directory as the json file, with the same name but with
            the extension *.nc. This is the default saving behavior of Solution.to_files.
        """
        if nc_filepath is None:
            nc_filepath = _construct_nc_filepath_from_json_filepath(json_filepath)
        simulation_result = xa.open_dataset(nc_filepath, engine='h5netcdf').load()
        simulation_result.close() # this is needed to release the lock on the file so that it can be written to again
        return Solution.from_json(
            json_string = json_filepath.read_text(),
            simulation_result = simulation_result,
        )



from __future__ import annotations

import argparse
import json
import logging
import os
import random
import signal
import sys
import time
from contextlib import suppress
from typing import Any, Dict

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_comms.py
# Import LIFU modules
from openlifu.io.LIFUInterface import LIFUInterface

# Set up logging with timestamps
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s'
)

# ---- Globals ----
CURRENT_IFC: dict[str, LIFUInterface | None] = {"iface": None}

# Signal handling for graceful exit
def signal_handler(_signum: int, _frame: Any) -> None:
    print("\n\n🛑 Stopping test gracefully...")
    try:
        if CURRENT_IFC["iface"] is not None:
            CURRENT_IFC["iface"].close()
            logging.info("Interface safely closed.")
    except Exception as e:
        logging.warning(f"Error closing interface: {e}")
    finally:
        CURRENT_IFC["iface"] = None
    sys.exit(0)

signal.signal(signal.SIGINT, signal_handler)

# Constants for test validation
EXPECTED_VERSION = ["v2.0.1", "v1.0.14"]
EXPECTED_TX7332_COUNT = 2
TEMP_MIN = 20.0
TEMP_MAX = 50.0
FREQ_MIN = 10
FREQ_MAX = 40
PULSE_WIDTH_MIN = 500
PULSE_WIDTH_MAX = 5000
PULSE_INTERVAL_MIN = 125000
PULSE_INTERVAL_MAX = 500000
PULSE_COUNT_MIN = 1
PULSE_COUNT_MAX = 10
PULSE_TRAIN_COUNT_MIN = 1
PULSE_TRAIN_COUNT_MAX = 10

# Trigger modes (use actual values from your code)
TRIGGER_MODE_SEQUENCE = 1
TRIGGER_MODE_CONTINUOUS = 2
TRIGGER_MODE_SINGLE = 3

# Test parameters
TEST_ITERATIONS = int(os.getenv("TEST_ITERATIONS", "5"))
TIMEOUT = int(os.getenv("TEST_TIMEOUT", "3"))

# Helper: Safe execution (no concurrency). Logs duration and errors, returns None on failure.
def safe_execute(func, description: str, *args, **kwargs) -> Any:
    start = time.time()
    try:
        result = func(*args, **kwargs)
        elapsed = time.time() - start
        if elapsed > TIMEOUT:
            logging.warning(f"⚠ '{description}' took {elapsed:.2f}s (> {TIMEOUT}s)")
        else:
            logging.info(f"✅ '{description}' completed in {elapsed:.2f}s")
        return result
    except Exception as e:
        logging.error(f"❌ Failed to {description}: {e}")
        return None

# Helper: Check if value is in valid range
def is_in_range(value: float, min_val: float, max_val: float) -> bool:
    return min_val <= value <= max_val

# Helper: Generate random valid trigger parameters
def generate_random_trigger_params() -> Dict[str, Any]:
    return {
        "TriggerFrequencyHz": random.randint(FREQ_MIN, FREQ_MAX),
        "TriggerPulseCount": random.randint(PULSE_COUNT_MIN, PULSE_COUNT_MAX),
        "TriggerPulseWidthUsec": random.randint(PULSE_WIDTH_MIN, PULSE_WIDTH_MAX),
        "TriggerPulseTrainInterval": random.randint(PULSE_INTERVAL_MIN, PULSE_INTERVAL_MAX),
        "TriggerPulseTrainCount": random.randint(PULSE_TRAIN_COUNT_MIN, PULSE_TRAIN_COUNT_MAX),
        "TriggerMode": random.choice([TRIGGER_MODE_SEQUENCE, TRIGGER_MODE_CONTINUOUS, TRIGGER_MODE_SINGLE]),
        "ProfileIndex": 0,
        "ProfileIncrement": 0
    }

# Helper: Compare two dicts (allowing small float differences)
def dicts_equal_with_tolerance(d1: Dict, d2: Dict, tolerance: float = 1e-6) -> bool:
    if set(d1.keys()) != set(d2.keys()):
        return False
    for k, v1 in d1.items():
        v2 = d2.get(k)
        if isinstance(v1, float) and isinstance(v2, float):
            if abs(v1 - v2) > tolerance:
                return False
        elif v1 != v2:
            return False
    return True
def _as_int(x):
    try:
        return int(x)
    except Exception:
        return x

def _as_float(x):
    try:
        return float(x)
    except Exception:
        return x

# Compare only the fields we set, allow device-side coercions and small float diffs
def trigger_dict_matches_set(set_params: Dict[str, Any],
                             reply: Dict[str, Any],
                             tol: float = 1e-6) -> bool:
    if not isinstance(reply, dict):
        return False
    keys_to_check = [
        "TriggerFrequencyHz",
        "TriggerPulseCount",
        "TriggerPulseWidthUsec",
        "TriggerPulseTrainInterval",
        "TriggerPulseTrainCount",
        "TriggerMode",
        "ProfileIndex",
        "ProfileIncrement",
    ]
    for k in keys_to_check:
        v_set = set_params.get(k)
        v_rep = reply.get(k)
        # allow numeric type coercions
        if isinstance(v_set, float) or isinstance(v_rep, float):
            if abs(_as_float(v_set) - _as_float(v_rep)) > tol:
                return False
        elif _as_int(v_set) != _as_int(v_rep):
            return False
    return True

# Main test function
def run_tx_test() -> int:
    print("🚀 Starting LIFU Transmitter Module Test Script...")
    test_results = {"total": 0, "passed": 0, "failed": 0}

    try:
        # Initialize interface
        CURRENT_IFC["iface"] = LIFUInterface()
        tx_connected, hv_connected = CURRENT_IFC["iface"].is_device_connected()
        if not tx_connected and hv_connected:
            print("❌ LIFU Device not fully connected. Attempting to power 12V...")
            with suppress(Exception):
                CURRENT_IFC["iface"].hvcontroller.turn_12v_on()
            time.sleep(2)

            # If there's monitoring that interferes with reconnection, stop it (ignore if missing)
            with suppress(Exception):
                CURRENT_IFC["iface"].stop_monitoring()

            # Re-init
            with suppress(Exception):
                CURRENT_IFC["iface"].close()
            CURRENT_IFC["iface"] = None
            time.sleep(3)
            print("🔄 Reinitializing after 12V power-on...")
            CURRENT_IFC["iface"] = LIFUInterface()
            tx_connected, hv_connected = CURRENT_IFC["iface"].is_device_connected()
            if not tx_connected or not hv_connected:
                print(f"❌ Still not connected. TX: {tx_connected}, HV: {hv_connected}")
                return 1

        if not tx_connected:
            print("❌ TX device not connected. Cannot proceed with tests.")
            return 1

        print("✅ LIFU Transmitter Device connected.")

        # Basic connectivity tests
        print(f"\n🔍 Testing: Transmitter (iterating {TEST_ITERATIONS} times)...")
        for i in range(TEST_ITERATIONS):
            print(f"  ➤ Iteration {i+1}/{TEST_ITERATIONS}...")
            params = generate_random_trigger_params()

            # Test 1: Ping
            print("\n🔍 Testing: Ping...")
            ping_result = safe_execute(CURRENT_IFC["iface"].txdevice.ping, "ping")
            if ping_result is True:
                print("✅ Ping successful (expected: always True)")
                test_results["passed"] += 1
            else:
                print("❌ Ping failed (should always return True)")
                test_results["failed"] += 1
            test_results["total"] += 1

            # Test 2: Get Version
            print("\n🔍 Testing: Get Version...")
            version = safe_execute(CURRENT_IFC["iface"].txdevice.get_version, "get version")
            if version in EXPECTED_VERSION:
                print(f"✅ Version correct: {version}")
                test_results["passed"] += 1
            else:
                print(f"❌ Version mismatch: expected '{EXPECTED_VERSION}', got '{version}'")
                test_results["failed"] += 1
            test_results["total"] += 1

            # Test 3: Enumerate TX7332 Devices
            print("\n🔍 Testing: Enumerate TX7332 Devices...")
            num_devices = safe_execute(CURRENT_IFC["iface"].txdevice.enum_tx7332_devices, "enum devices")
            if num_devices == EXPECTED_TX7332_COUNT:
                print(f"✅ Found exactly {num_devices} TX7332 devices (expected: {EXPECTED_TX7332_COUNT})")
                test_results["passed"] += 1
            else:
                print(f"❌ Wrong number of devices: expected {EXPECTED_TX7332_COUNT}, got {num_devices}")
                test_results["failed"] += 1
            test_results["total"] += 1

            # Test 4: Temperature
            print("\n🔍 Testing: Temperature...")
            temp = safe_execute(CURRENT_IFC["iface"].txdevice.get_temperature, "get temperature")
            if temp is not None and is_in_range(temp, TEMP_MIN, TEMP_MAX):
                print(f"✅ Temperature: {temp:.1f}°C (in valid range: {TEMP_MIN}-{TEMP_MAX}°C)")
                test_results["passed"] += 1
            else:
                print(f"❌ Temperature out of range: {temp}°C (expected: {TEMP_MIN}-{TEMP_MAX}°C)")
                test_results["failed"] += 1
            test_results["total"] += 1

            # Test 5: Set & Get Trigger (loop over iterations)
            print(f"    Setting trigger: {json.dumps(params, indent=2)}")
            try:
                set_reply = safe_execute(
                    CURRENT_IFC["iface"].txdevice.set_trigger_json,
                    "set trigger",
                    data=params
                )
            except TypeError:
                # If API expects a positional dict
                set_reply = safe_execute(lambda p=params: CURRENT_IFC["iface"].txdevice.set_trigger_json(p), "set trigger")

            # Interpret success: dict reply means success; True also okay (legacy)
            if isinstance(set_reply, dict) or set_reply is True:
                if isinstance(set_reply, dict):
                    if trigger_dict_matches_set(params, set_reply):
                        print("    ✅ Trigger set successfully (echo validated).")
                    else:
                        print("    ⚠ Trigger set returned but values differ after device coercion:")
                        print("       ", json.dumps(set_reply, indent=2))
                else:
                    print("    ✅ Trigger set successfully.")
            else:
                print(f"    ❌ Failed to set trigger: {set_reply}")
                test_results["failed"] += 1
                test_results["total"] += 1
                continue  # skip read-back on failure


            # Read back trigger
            get_result = safe_execute(CURRENT_IFC["iface"].txdevice.get_trigger_json, "get trigger")
            if get_result is None:
                print("    ❌ Failed to read trigger after set.")
                test_results["failed"] += 1
            else:
                print(f"    Read back: {json.dumps(get_result, indent=2)}")

                # Validate all fields
                passed = True
                for key in params:
                    if key == "TriggerMode":
                        if get_result.get(key) != params[key]:
                            print(f"    ❌ TriggerMode mismatch: expected {params[key]}, got {get_result.get(key)}")
                            passed = False
                    elif key in [
                        "TriggerFrequencyHz", "TriggerPulseCount", "TriggerPulseWidthUsec",
                        "TriggerPulseTrainInterval", "TriggerPulseTrainCount"
                    ]:
                        val1 = params[key]
                        val2 = get_result.get(key)
                        if val2 is None or abs(float(val1) - float(val2)) > 1e-6:
                            print(f"    ❌ {key} mismatch: expected {val1}, got {val2}")
                            passed = False
                    else:
                        # Skip profile fields for now
                        continue

                if passed:
                    print("    ✅ All trigger values match!")
                    test_results["passed"] += 1
                else:
                    test_results["failed"] += 1
            test_results["total"] += 1

        # Final Summary
        print("\n" + "="*70)
        print("📊 TEST SUMMARY")
        print(f"Total Tests: {test_results['total']}")
        print(f"Passed:      {test_results['passed']}")
        print(f"Failed:      {test_results['failed']}")
        print("="*70)

        if test_results["failed"] == 0:
            print("🎉 ALL TESTS PASSED!")
            return 0
        else:
            print("❌ Some tests failed.")
            return 1

    except Exception as e:
        logging.critical(f"Unexpected error during test: {e}")
        return 1

    finally:
        if CURRENT_IFC["iface"] is not None:
            with suppress(Exception):
                CURRENT_IFC["iface"].close()
                logging.info("Interface safely closed.")
            CURRENT_IFC["iface"] = None

if __name__ == "__main__":
    # Argument parsing
    parser = argparse.ArgumentParser(description="Test LIFU Transmitter Module")
    parser.add_argument('--iterations', type=int, default=5, help='Number of trigger test iterations')
    parser.add_argument('--timeout', type=int, default=3, help='Soft timeout (seconds) for warning logs')
    args = parser.parse_args()

    # Override defaults from environment or args
    TEST_ITERATIONS = args.iterations
    TIMEOUT = args.timeout

    # Run test
    sys.exit(run_tx_test())
